﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RidgeDetector.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.AbstractedLines
{
	public enum BlendMode
	{
		Additive    = 0,
		Multicative = 1,
	}

	public class RidgeDetector : FboNode
	{
		#region Properties
//=================================================================================================

		[Uniform(), Category("Ridge detection"), DisplayName("Line width"), Description("Width of lines to draw, in pixels.")]
		public float LineWidth { get; set; }

		[Uniform("steps"), Category("Ridge detection"), DisplayName("Steps"), DefaultValue(5), Description("The maximum number of steps towards the ridge or valley.")]
		public int NumSteps { get; set; }

		[Uniform(), Category("Ridge detection"), Description("Fraction of the full distance to move towards line.")]
		public float Beta { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Valleys lower threshold"), Description("Lower threshold on curvature for dark valleys.")]
		public float ValleysLower { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Valleys upper threshold"), Description("Upper threshold on curvature for dark valleys.")]
		public float ValleysUpper { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Valleys alpha")]
		public float ValleysAlpha { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (diffuse) lower threshold"), Description("Lower threshold on curvature for light ridges in the diffuse tone image.")]
		public float RidgesDLower { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (diffuse) upper threshold"), Description("Upper threshold on curvature for light ridges in the diffuse tone image.")]
		public float RidgesDUpper { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (diffuse) alpha")]
		public float RidgesDAlpha { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (specular) lower threshold"), Description("Lower threshold on curvature for light ridges in the specular tone image.")]
		public float RidgesSLower { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (specular) upper threshold"), Description("Upper threshold on curvature for light ridges in the specular tone image.")]
		public float RidgesSUpper { get; set; }

		[Uniform(), Category("Ridge detection"), DisplayName("Ridges (specular) alpha")]
		public float RidgesSAlpha { get; set; }

		[Uniform(), Category("Output"), DisplayName("Blend mode"), Description("How to blend toon and video colours.")]
		public BlendMode BlendMode { get; set; }

		[Uniform(), Category("Output"), DisplayName("Blend factor"), Description("Factor for blending between input video (0) and toon look (1).")]
		public float BlendFactor { get; set; }

		[Uniform(), Category("Output"), DefaultValue(1.0f), Description("Saturation of the video colour.")]
		public float Saturation { get; set; }

//=================================================================================================
		#endregion // Properties

		public RidgeDetector()
		{
			// default settings
			LineWidth = 10;
			NumSteps = 5;
			Beta = 0.3f;

			RidgesDLower = 0.005f;
			RidgesDUpper = 0.015f;
			RidgesSLower = 0.005f;
			RidgesSUpper = 0.02f;
			ValleysLower = 0.001f;
			ValleysUpper = 0.03f;
			RidgesDAlpha = 1;
			RidgesSAlpha = 1;
			ValleysAlpha = 1;

			BlendMode   = BlendMode.Multicative;
			BlendFactor = 0.5f;
			Saturation  = 1.3f;

			AddDefaultPins();
			AddInputPin("video");
		}

		#region Rendering functions
//=================================================================================================

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("AbstractedLines.Shaders.RidgeDetector.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in",    Uniform = "input", Unit = 0 },
				new TextureBinding { Pin = "video", Uniform = "video", Unit = 1 },
			};

			base.Init();
		}

		public override TextureFrame Draw(Dictionary<string, TextureFrame> textures)
		{
			// create output texture to fit input texture
			var resultTexture = TextureFrame.CreateToMatch(textures["in"]);
			resultTexture.Alloc();

			Draw(textures, resultTexture);

			return resultTexture;
		}

//=================================================================================================
		#endregion // Rendering function
	}
}
