﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RidgeDetector.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input tone images
uniform sampler2D input;

// input video image
uniform sampler2D video;

// "Width of lines to sraw, in pixels."
uniform float lineWidth;

// The maximum number of steps towards the ridge or valley.
uniform int steps;

// Fraction of the full distance to move towards line.
uniform float beta;

// Lower thresholds on curvature.
uniform float valleysLower;
uniform float ridgesDLower;
uniform float ridgesSLower;

// Upper thresholds on curvature.
uniform float valleysUpper;
uniform float ridgesDUpper;
uniform float ridgesSUpper;

// alphas of lines
uniform float valleysAlpha;
uniform float ridgesDAlpha;
uniform float ridgesSAlpha;

// factor for blending between video (0) and toon look (1)
uniform int   blendMode;
uniform float blendFactor;
uniform float saturation;

// Reads in tone values for the given channel and pixel position
mat3 readToneImage(sampler2D image, int channel, vec2 position)
{
	// read in tone values
	vec2 texelSize = 1.0 / textureSize2D(image, 0).xy;
	vec3 t1 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1., -1.)) * texelSize).rgb;
	vec3 t2 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0., -1.)) * texelSize).rgb;
	vec3 t3 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1., -1.)) * texelSize).rgb;
	vec3 t4 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1.,  0.)) * texelSize).rgb;
	vec3 t5 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0.,  0.)) * texelSize).rgb;
	vec3 t6 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1.,  0.)) * texelSize).rgb;
	vec3 t7 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1.,  1.)) * texelSize).rgb;
	vec3 t8 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0.,  1.)) * texelSize).rgb;
	vec3 t9 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1.,  1.)) * texelSize).rgb;

	switch(channel)
	{
		case 0: return mat3(t1.x, t2.x, t3.x, t4.x, t5.x, t6.x, t7.x, t8.x, t9.x);
		case 1: return mat3(t1.y, t2.y, t3.y, t4.y, t5.y, t6.y, t7.y, t8.y, t9.y);
		case 2: return mat3(t1.z, t2.z, t3.z, t4.z, t5.z, t6.z, t7.z, t8.z, t9.z);
	}
}

// Calculates line opacity as close to Lee et al. 2007 as possible.
float calcLineOpacity(sampler2D image, int channel, vec2 position, bool highlights, float c_l, float c_u)
{
	// position ucrrently under consideration
	vec2 pos = position;

	// distance 'travelled' so far
	float dist = 0;

	for(int i = 0; i < steps; i++)
	{
		mat3 t = readToneImage(image, channel, pos);
		if(!highlights) t = 1 - t; // invert tone image to detect valleys
		
		// compute coefficents of degree-2 polynomial
		//
		//   f(x,y) = a0 * x^2 + 2 * a1 * x * y + a2 * y^2 + a3 * x + a4 * y + c5
		//
		// using A = H * T (as described in Lee et al. 2007), where H is hardcoded
		// with values determined in Matlab:
		//
		//   [x,y] = meshgrid(-1:1)
		//   y = y'; y = y(:); x = x'; x = x(:);
		//   X = [x .^ 2, 2 * x .* y, y .^ 2, x, y, ones(size(x))]
		//   H = inv(X' * X) * X'
		//   norm(H - [[1 -2 1 1 -2 1 1 -2 1]/6; [1 0 -1 0 0 0 -1 0 1]/8; [1 1 1 -2 -2 -2 1 1 1]/6; ...
		//       [-1 0 1 -1 0 1 -1 0 1]/6; [-1 -1 -1 0 0 0 1 1 1]/6; [-1 2 -1 2 5 2 -1 2 -1]/9])
		//   >> 1.4007e-016
		float invLineWidth = 1.0 / lineWidth;
		float a0 = ( t[0][0] - 2.0 * t[0][1] + t[0][2] +       t[1][0] - 2.0 * t[1][1] +       t[1][2] + t[2][0] - 2.0 * t[2][1] + t[2][2]) / 6.0;// * invLineWidth * invLineWidth;
		float a1 = ( t[0][0]                 - t[0][2]                                                 - t[2][0]                 + t[2][2]) / 8.0;// * invLineWidth * invLineWidth;
		float a2 = ( t[0][0] +       t[0][1] + t[0][2] - 2.0 * t[1][0] - 2.0 * t[1][1] - 2.0 * t[1][2] + t[2][0] +       t[2][1] + t[2][2]) / 6.0;// * invLineWidth * invLineWidth;
		float a3 = (-t[0][0]                 + t[0][2] -       t[1][0] +                       t[1][2] - t[2][0]                 + t[2][2]) / 6.0;// * invLineWidth;
		float a4 = (-t[0][0] -       t[0][1] - t[0][2]                                 +       t[2][0]           +       t[2][1] + t[2][2]) / 6.0;// * invLineWidth;
		float a5 = (-t[0][0] + 2.0 * t[0][1] - t[0][2] + 2.0 * t[1][0] + 5.0 * t[1][1] + 2.0 * t[1][2] - t[2][0] + 2.0 * t[2][1] - t[2][2]) / 9.0;

		// compute eigenvalues of M = [a0 a1; a1 a2] (see for details:
		// <http://www.math.harvard.edu/archive/21b_fall_04/exhibits/2dmatrices/index.html>)
		// Note: adding small epsilon to work around machine precision (sqrt(negative) = NaN).
		float eigv_term = sqrt(pow((a0 + a2), 2.0) / 4.0 - (a0 * a2 - a1 * a1) + 1e-10);
		float l1 = (a0 + a2) / 2.0 + eigv_term;
		float l2 = (a0 + a2) / 2.0 - eigv_term;
				
		// reject if the first principal curvature is too low (i.e. locally flat)
		float c = max(abs(l1), abs(l2));
		if(c < c_l) return 0.0;
		
		// compute eigenvectors of M, exploiting that M is symmetric.
		vec2 v1, v2;
		if(abs(a1) < 1.e-10) { v1 = vec2(1, 0); v2 = vec2(0, 1); }
		else { v1 = normalize(vec2(l1 - a2, a1)); v2 = normalize(vec2(l2 - a2, a1)); }

		// compute location of ridge relative to current position
		vec2 ridge;
		if(abs(a0 * a2 - a1 * a1) < 1.e-10) // M is singular
		{
			ridge = -dot(vec2(a3, a4), v1) / (2 * l1) * v1; // l2 is zero, so v2 is 'u' in the paper
		}
		else
		{
			ridge = - 0.5 / (a0 * a2 - a1 * a1) * vec2((a2 * a3 - a1 * a4), (a0 * a4 - a1 * a3));
			ridge -= dot(ridge, (abs(l1) > abs(l2) ? v2 : v1)) * (abs(l1) > abs(l2) ? v2 : v1);
		}
		
		ridge *= lineWidth;
		
		// reject if too far away from ridge
		if(length(ridge) > lineWidth) return 0.0;
		
		// reject if travelled too far
		dist += length(beta * ridge);
		if(2 * dist > lineWidth) 0.0;
				
		// made it to the end
		if(i == steps - 1)
		{
			// reject pixels in valleys
			if(abs(l1) > abs(l2) && l1 > 0) return 0.0;

			// accept pixel
			float opacity = 1;
			
			// apply soft threshold
			opacity *= clamp((c - c_l) / (c_u - c_l), 0, 1);
			
			// apply distance fall-off
			opacity *= clamp(0.75 - length(gl_FragCoord.xy - pos - ridge) / lineWidth, 0, 1);
			
			// apply "counterchange"
			float tone;
			switch(channel)
			{
				case 0: tone = texture2D(image, gl_FragCoord.xy / textureSize2D(image, 0).xy).x; break;
				case 1: tone = texture2D(image, gl_FragCoord.xy / textureSize2D(image, 0).xy).y; break;
				case 2: tone = texture2D(image, gl_FragCoord.xy / textureSize2D(image, 0).xy).z; break;
			}
			opacity *= highlights ? tone : 1 - tone;

			return opacity;
		}

		// walk towards the ridge
		pos  += beta * ridge;
	}

	// fall through should not happen
	return 0.0;
}

void main(void)
{
	// compute opacities for all line types
	float valleys = calcLineOpacity(input, 0, gl_FragCoord.xy, false, valleysLower, valleysUpper);
	float ridgesD = calcLineOpacity(input, 0, gl_FragCoord.xy, true,  ridgesDLower, ridgesDUpper);
	float ridgesS = calcLineOpacity(input, 1, gl_FragCoord.xy, true,  ridgesSLower, ridgesSUpper);
	
	// pick base coat
	//vec3 colour = texture2D(input, gl_FragCoord.xy / textureSize2D(input, 0).xy).xxx; // diffuse tone image
	//vec3 colour = texture2D(input, gl_FragCoord.xy / textureSize2D(input, 0).xy).yyy; // specular tone image
	vec3 colour = texture2D(input, gl_FragCoord.xy / textureSize2D(input, 0).xy).zzz; // toon shaded
	//vec3 colour = vec3(0.7); // grey
	
	// read in video pixel
	vec3 videocolour = texture2D(video, gl_FragCoord.xy / textureSize2D(video, 0)).rgb;
	
	// change saturation of video colour
	vec3 videogrey = vec3(dot(videocolour, vec3(0.2126729, 0.7151522, 0.0721750)));
	videocolour = mix(videogrey, videocolour, saturation);
	videocolour = clamp(videocolour, 0.0, 1.0);
	
	switch(blendMode)
	{
		case 0: // additive
			colour = mix(videocolour, colour, blendFactor);
			break;
			
		case 1: // multiplicative
			colour = clamp(pow(colour, vec3(blendFactor)) * videocolour, 0.0, 1.0);
			break;
	}

	// blend in the lines
	colour = mix(colour, vec3(1), ridgesDAlpha * ridgesD);
	colour = mix(colour, vec3(1), ridgesSAlpha * ridgesS);
	colour = mix(colour, vec3(0), valleysAlpha * valleys);

	// blend between video and toon
	gl_FragColor = vec4(colour, 1);
}