﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: FillInDownsamplingFilter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.FillIn
{
	/// <summary>Node that downsamples an input texture using nearest neighbours.</summary>
	public class FillInDownsamplingFilter : FboNode
	{
		/// <summary>Gets or sets the downsampling factor.</summary>
		[Uniform(), Description("Downsampling factor.")]
		public int Factor { get; set; }

		/// <summary>Gets or sets the sample offset in the input image. Default: 0.</summary>
		[Uniform(), DefaultValue(0), Description("The offset of the sample in the input image.")]
		public int SampleOffset { get; set; }

		/// <summary>Initialises a new instance of the <see cref="FillInDownsamplingFilter"/> class.</summary>
		public FillInDownsamplingFilter()
		{
			// default settings
			Factor = 2;
			SampleOffset = 0;

			AddDefaultPins();
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("FillIn.Shaders.FillInDownsamplingFilter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "input", Unit = 0 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate the output texture of the new smaller size, but otherwise identical to the input texture
			var outputWidth  = (inputs["in"].Width  - SampleOffset + Factor - 1) / Factor;
			var outputHeight = (inputs["in"].Height - SampleOffset + Factor - 1) / Factor;
			var resultTexture = new TextureFrame(outputWidth, outputHeight)
			{
				SequenceNumber = inputs["in"].SequenceNumber,
				InternalFormat = inputs["in"].InternalFormat,
				MinFilter      = inputs["in"].MinFilter,
				MagFilter      = inputs["in"].MagFilter,
				WrapS          = inputs["in"].WrapS,
				WrapT          = inputs["in"].WrapT
			};
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
