﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RenderQuads.geom 537 2012-04-04 15:35:42Z cr333 $
*/
#version 130 
#extension GL_EXT_geometry_shader4 : enable

uniform vec4 backgroundPlane;
uniform float planeThreshold;
uniform float pointSize;
uniform float attenuationFactor;
uniform vec3 lightDirection;
uniform float jitterPosition;
uniform float jitterRotation;

in vec3 vertex[];
in vec3 normal[];
in vec2 direction[];

out vec2 coords;
out float attenuation;

// source: <http://www.geeks3d.com/20100831/shader-library-noise-and-pseudo-random-number-generator-in-glsl/>
float rnd(vec2 x)
{
	int n = int(x.x * 40.0 + x.y * 6400.0);
	n = (n << 13) ^ n;
	return 1.0 - float( (n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0;
}

void main(void)
{
	for(int i = 0; i < gl_VerticesIn; i++)
	{
		vec4 centre = gl_PositionIn[i];
		gl_FrontColor = gl_FrontColorIn[i];
		//gl_FrontColor = vec4(normalize(normal[i]) / 2 + 0.5, 1);
		//attenuation = clamp(1.0 - attenuationFactor * abs(dot(normalize(normal[i]), normalize(lightDirection))), 0.0, 1.0); // 1 - k * diffuse
		attenuation = clamp(1.0 - attenuationFactor * pow(abs(dot(normalize(normal[i]), normalize(lightDirection))), 0.5), 0.0, 1.0); // 1 - k * sqrt(diffuse)
		if(isnan(attenuation)) attenuation = 0.0;
		
		//// screen-aligned quad
		//gl_Position = gl_ModelViewProjectionMatrix * (centre + pointSize * vec4(-0.5, -0.5, 0, 0)); EmitVertex();
		//gl_Position = gl_ModelViewProjectionMatrix * (centre + pointSize * vec4(-0.5,  0.5, 0, 0)); EmitVertex();
		//gl_Position = gl_ModelViewProjectionMatrix * (centre + pointSize * vec4( 0.5, -0.5, 0, 0)); EmitVertex();
		//gl_Position = gl_ModelViewProjectionMatrix * (centre + pointSize * vec4( 0.5,  0.5, 0, 0)); EmitVertex();
		//EndPrimitive();
		
		//// screen-aligned quad
		//vec3 u = vec3(1,0,0);
		//vec3 v = vec3(0,1,0);
		
		//// oriented using x axis and normal
		//vec3 n = normalize(normal[i]);
		//vec3 v = normalize(cross(n, vec3(1,0,0))) / 3;
		//vec3 u = normalize(cross(v, n));
		
		// normalise plane normal (so that the distance to the plane is meaningful)
		vec4 plane = backgroundPlane / length(backgroundPlane.xyz);
		
		// discard sprites close to the background plane
		if(dot(vec4(vertex[i], 1.0), backgroundPlane) < planeThreshold)
			continue;
		
		// oriented along principal curvature(?)
		vec3 n = normalize(normal[i]);
		//vec3 n = vec3(0,0,-1);
		vec3 v = normalize(cross(n, vec3(direction[i], 0)));
		vec3 u = normalize(cross(v, n));
		v = normalize(cross(u, vec3(0,0,-1))) / 5.0f;
		
		// jitter position
		vec4 jitterPos = jitterPosition * vec4(2.0 * rnd(centre.xy) - 1.0, 2.0 * rnd(centre.yx) - 1.0, 0.0, 0.0);
		
		// jitter rotation
		float jitterAngle = radians(jitterRotation) * (2.0 * rnd(centre.xy) - 1.0);
		mat3 jitterRot = mat3(cos(jitterAngle), sin(jitterAngle), 0.0, -sin(jitterAngle), cos(jitterAngle), 0.0, 0.0, 0.0, 1.0);
		u = jitterRot * u;
		v = jitterRot * v;
		
		// quad orthogonal to normal
		vec4 position = (centre + jitterPos + pointSize * vec4(-u-v, 0) / 2.0); position.z = vertex[i].z;
		gl_Position = gl_ModelViewProjectionMatrix * position;
		coords = vec2(-1.0, -1.0);
		EmitVertex();
		
		position = (centre + jitterPos + pointSize * vec4(-u+v, 0) / 2.0); position.z = vertex[i].z;
		gl_Position = gl_ModelViewProjectionMatrix * position;
		coords = vec2(-1.0, 1.0);
		EmitVertex();
		
		position = (centre + jitterPos + pointSize * vec4(+u-v, 0) / 2.0); position.z = vertex[i].z;
		gl_Position = gl_ModelViewProjectionMatrix * position;
		coords = vec2(1.0, -1.0);
		EmitVertex();
		
		position = (centre + jitterPos + pointSize * vec4(+u+v, 0) / 2.0); position.z = vertex[i].z;
		gl_Position = gl_ModelViewProjectionMatrix * position;
		coords = vec2(1.0, 1.0);
		EmitVertex();
	}
	
}