﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: ReprojectionNode.cs 556 2012-09-12 11:31:49Z cr333 $
*/
using System;
using System.Linq;
using System.Diagnostics;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib;
using VideoLib.Frames;
using VideoLib.OpenGL;
using VideoLib.Vision;

namespace RGBZcamera.Library
{
	public enum DepthCleanupMode
	{
		Disabled                = 0,
		SobelThresholding       = 1,
		FaceNormalThresholding  = 2,
		PixelNormalThresholding = 3,
	}

	/// <summary>Implements a OpenGL-based mesh reprojection node.</summary>
	public partial class ReprojectionNode : Node, IOpenGLNode
	{
		/// <summary>Gets or sets the extrinsic parameters of the source camera (i.e. ToF).</summary>
		public ExtrinsicParameters SourceExtrinsics { get; set; }

		/// <summary>Gets or sets the intrinsic parameters of the source camera (i.e. ToF).</summary>
		public IntrinsicParameters SourceIntrinsics { get; set; }

		/// <summary>Gets or sets the extrinsic parameters of the target camera (i.e. video camera).</summary>
		public ExtrinsicParameters TargetExtrinsics { get; set; }

		/// <summary>Gets or sets the target intrinsic parameters of the target camera (i.e. video camera).</summary>
		public IntrinsicParameters TargetIntrinsics { get; set; }

		/// <summary>Gets or sets a value indicating whether the view is cropped. Default: true.</summary>
		public bool CropView { get; set; }

		/// <summary>Gets or sets the near plane used in the cropping frame computation. Default: 0.5 (m).</summary>
		public float CropNear { get; set; }

		/// <summary>Gets or sets the farplane used in the cropping frame computation. Default: 5.0 (m).</summary>
		public float CropFar { get; set; }

		/// <summary>Gets or sets a value indicating the base output resolution (true=video; false=depth). Default: true (video resolution).</summary>
		public bool OutputVideoResolution { get; set; }

		public float SobelThreshold { get; set; }

		public float NormalThreshold { get; set; }

		public DepthCleanupMode CleanupMode { get; set; }

		/// <summary>Constructs an OpenGL-based mesh reprojection node.</summary>
		public ReprojectionNode()
		{
			OutputVideoResolution = true;
			SobelThreshold        = 0.5f;
			NormalThreshold       = 0.5f;
			CleanupMode           = DepthCleanupMode.FaceNormalThresholding;
			CropView              = true;
			CropNear              = 0.5f;
			CropFar               = 5.0f;

			AddInputPins( "pointsIn",  "colourIn");
			AddOutputPins("pointsOut", "colourOut");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			GL.GenBuffers(1, out VertexBufferID);
			GL.GenBuffers(1, out IndexBufferID);

			SetupDepthShaders();
		}

		public override void Process()
		{
			// get latest frames
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textureFrames = CastFramesTo<TextureFrame>(frames);

			var pointsFrame = textureFrames["pointsIn"];
			var colourFrame = textureFrames["colourIn"];
			if(OutputVideoResolution) VideoSize = new Size(colourFrame.Width, colourFrame.Height);
			else                      VideoSize = new Size(pointsFrame.Width, pointsFrame.Height);

			// update mesh
			MeshTools.UpdateBaseMeshVertexBuffer(pointsFrame.Width, pointsFrame.Height, VertexBufferID, out NumVertices);
			MeshTools.UpdateBaseMeshIndexBuffer( pointsFrame.Width, pointsFrame.Height, IndexBufferID,  out NumIndices);

			ComputeCropFrame(pointsFrame);
			SetupFramebufferObject();

			var colourTex = DrawVideoTexture(colourFrame);
			PushFrame("colourOut", colourTex);
			colourTex.RefDec();

			var pointsTex = DrawMesh(pointsFrame);
			PushFrame("pointsOut", pointsTex);
			pointsTex.RefDec();

			// release input textures
			textureFrames.RefDec();
		}

		#region Cropping computations
//=================================================================================================

		/// <summary>Cropping region relative to video frame (OpenGL: x is right, y is up).</summary>
		private Rectangle Crop;
		private Size CropSize;

		private void ComputeCropFrame(VideoFrame pointCloud)
		{
			// check if we need to recompute the crop rectangle
			if (pointCloud.Width == CropSize.Width && pointCloud.Height == CropSize.Height)
				return;

			// check that we have all parameters
			if (SourceExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of source camera not provided.");
			if (SourceIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of source camera not provided.");
			if (TargetExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of target camera not provided.");
			if (TargetIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of target camera not provided.");

			//// Detailed debug print of all used variables.
			//Console.WriteLine("===============================================================================");
			//Console.WriteLine("SourceExtrinsics: [{0}]", SourceExtrinsics.P.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine("SourceIntrinsics:");
			//Console.WriteLine("  F = [{0}, {1}]", SourceIntrinsics.Fx, SourceIntrinsics.Fy);
			//Console.WriteLine("  C = [{0}]", SourceIntrinsics.C.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine("  K = [{0}]", SourceIntrinsics.K.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine("TargetExtrinsics: [{0}]", TargetExtrinsics.P.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine("TargetIntrinsics:");
			//Console.WriteLine("  F = [{0}, {1}]", TargetIntrinsics.Fx, TargetIntrinsics.Fy);
			//Console.WriteLine("  C = [{0}]", TargetIntrinsics.C.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine("  K = [{0}]", TargetIntrinsics.K.Aggregate("", (total, next) => total + ", " + next.ToString()));
			//Console.WriteLine();

			// This is how it works:
			// ---------------------
			// The cropping frame is determined by the frustum spanned by the time-of-flight sensor.
			// This is determined by back-projecting the outer corners of the depth image for both
			// the near (0.5 m) and the far planes (5 m), and then projecting them onto the target
			// camera's image plane. We then compute the largest axis-aligned rectangle that always
			// fits inside the projected mesh.

			// corner pixels of distance image, with minimum and maximum depth
			int xmin = 0; int xmax = pointCloud.Width  - 1;
			int ymin = 0; int ymax = pointCloud.Height - 1;
			var cornerPixels = new Vector3[] {
				new Vector3(xmin, ymin, CropNear), new Vector3(xmin, ymax, CropNear), new Vector3(xmax, ymin, CropNear), new Vector3(xmax, ymax, CropNear),
				new Vector3(xmin, ymin, CropFar ), new Vector3(xmin, ymax, CropFar ), new Vector3(xmax, ymin, CropFar ), new Vector3(xmax, ymax, CropFar ),
			};
			//Console.WriteLine("cornerPixels:");
			//Console.WriteLine("  [0] = {0}", cornerPixels[0]);
			//Console.WriteLine("  [1] = {0}", cornerPixels[1]);
			//Console.WriteLine("  [2] = {0}", cornerPixels[2]);
			//Console.WriteLine("  [3] = {0}", cornerPixels[3]);
			//Console.WriteLine();

			// compute coordinates of corners in source camera reference frame
			var corners = cornerPixels.Select(x => DistanceToPoint(x, SourceIntrinsics));
			//Console.WriteLine("corners:");
			//for (int i = 0; i < 4; i++) Console.WriteLine("  [{0}] = {1}", i, corners.ElementAt(i));
			//Console.WriteLine();

			// transform to target view and project
			var invSourceView = SourceExtrinsics.ToMatrix4(); invSourceView.Invert();
			var P = TargetIntrinsics.ToMatrix4() * TargetExtrinsics.ToMatrix4() * invSourceView; P.Transpose();
			var screenPoints = corners.Select(x => Vector4.Transform(x, P));
			//Console.WriteLine("P = {0}", P);
			//Console.WriteLine("screenPoints:");
			//for (int i = 0; i < 4; i++) Console.WriteLine("  [{0}] = {1}", i, screenPoints.ElementAt(i));
			//Console.WriteLine();

			// normalise to get screen coordinates
			var screenCoords = screenPoints.Select(x => new Vector2(x.X / x.Z, x.Y / x.Z));
			//Console.WriteLine("screenCoords:");
			//for (int i = 0; i < 4; i++) Console.WriteLine("  [{0}] = {1}", i, screenCoords.ElementAt(i));
			//Console.WriteLine();

			//// find the smallest rectangle that always contains the mesh (fit outer)
			//int cxmin = screenCoords.Select(x => (int)Math.Floor(x.X)).Min();
			//int cxmax = screenCoords.Select(x => (int)Math.Ceiling(x.X)).Max();
			//int cymin = screenCoords.Select(x => (int)Math.Floor(x.Y)).Min();
			//int cymax = screenCoords.Select(x => (int)Math.Ceiling(x.Y)).Max();

			// find the largest rectangle that always fits inside the mesh (fit inner)
			int cxmin = (int)Math.Floor(  new int[] { 0, 1, 4, 5 }.Select(i => screenCoords.ElementAt(i).X).Max());
			int cxmax = (int)Math.Ceiling(new int[] { 2, 3, 6, 7 }.Select(i => screenCoords.ElementAt(i).X).Min());
			int cymin = (int)Math.Floor(  new int[] { 0, 2, 4, 6 }.Select(i => screenCoords.ElementAt(i).Y).Max());
			int cymax = (int)Math.Ceiling(new int[] { 1, 3, 5, 7 }.Select(i => screenCoords.ElementAt(i).Y).Min());
			//Console.WriteLine("cxmin = {0}; cxmax = {1}; cymin = {2}; cymax = {3}", cxmin, cxmax, cymin, cymax);
			//Console.WriteLine();

			// flip Y for OpenGL
			int tmp = cymax;
			cymax = VideoSize.Height - cymin - 1;
			cymin = VideoSize.Height - tmp - 1;
			//Console.WriteLine("cxmin = {0}; cxmax = {1}; cymin = {2}; cymax = {3}  (after flip)", cxmin, cxmax, cymin, cymax);

			CropSize = new Size(pointCloud.Width, pointCloud.Height);
			if(CropView) Crop = new Rectangle(cxmin, cymin, cxmax - cxmin + 1, cymax - cymin + 1);
			else Crop = new Rectangle(0, 0, VideoSize.Width, VideoSize.Height); // no cropping
			//Console.WriteLine("CropSize = {0}", CropSize);
			//Console.WriteLine("Crop = {0}", Crop);
			//Console.WriteLine("===============================================================================");
		}

		/// <summary>Back-project point (pix_x, pix_y, distance) to a homogeneous 3D point.</summary>
		/// <param name="p">The p.</param>
		/// <returns></returns>
		private Vector4 DistanceToPoint(Vector3 p, IntrinsicParameters K)
		{
			// inv([Fx 0 Cx; 0 Fy Cy; 0 0 1]) = [iFx 0 iCx; 0 iFy iCy; 0 0 1]
			float iFx = 1.0f / K.Fx;
			float iFy = 1.0f / K.Fy;
			float iCx = -K.Cx / K.Fx;
			float iCy = -K.Cy / K.Fy;

			// inv(K) * [x y 1]' = [iFx 0 iCx; 0 iFy iCy; 0 0 1] * [x y 1]'
			float x3 = iFx * p.X + iCx;
			float y3 = iFy * p.Y + iCy;
			//float z3 = 1;

			// scale point to have desired distance along ray
			// p = p / norm(p) * depth
			float scale = p.Z / (float)Math.Sqrt(x3 * x3 + y3 * y3 + 1); // z3 = 1
			return new Vector4(x3 * scale, y3 * scale, scale, 1); // z3 = 1
		}
		
//=================================================================================================
		#endregion

		#region Framebuffer Object (off-screen rendering)
//=================================================================================================

		private int FramebufferID;
		private int DepthRenderbufferID;
		private Size FramebufferSize;

		private void SetupFramebufferObject()
		{
			// correct size already? nothing to do then ...
			if (Crop.Size == FramebufferSize) return;

			// delete existing buffers
			if (DepthRenderbufferID  != 0) GL.DeleteRenderbuffers(1, ref DepthRenderbufferID);
			if (FramebufferID        != 0) GL.DeleteFramebuffers( 1, ref FramebufferID);

			// generate a new depth renderbuffer
			GL.GenRenderbuffers(1, out DepthRenderbufferID);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, DepthRenderbufferID);
			GL.RenderbufferStorage(RenderbufferTarget.Renderbuffer, RenderbufferStorage.DepthComponent32f, Crop.Width, Crop.Height);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, 0);

			// generate a new framebuffer object
			GL.GenFramebuffers(1, out FramebufferID);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferRenderbuffer(FramebufferTarget.Framebuffer, FramebufferAttachment.DepthAttachment, RenderbufferTarget.Renderbuffer, DepthRenderbufferID);

			//Utils.CheckFboStatus(); // not complete, as colour buffer is missing

			FramebufferSize = Crop.Size;
		}
		
//=================================================================================================
		#endregion

		#region Depth shaders
//=================================================================================================

		private int MeshProgramID;

		private void SetupDepthShaders()
		{
			// load, compile and link shaders
			MeshProgramID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert"),
				Utils.CreateShaderFromResource("Shaders.Reprojection.vert"),
				Utils.CreateShaderFromResource("Shaders.Reprojection.frag")
			);

			// set uniforms
			GL.UseProgram(MeshProgramID);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "coordinates"), 0); // texture unit 0
			GL.UseProgram(0);
		}

//=================================================================================================
		#endregion

		#region Texture methods
//=================================================================================================

		/// <summary>The size of the input video frame.</summary>
		private Size VideoSize;

		private TextureFrame DrawVideoTexture(TextureFrame videoTexture)
		{
			// allocate the output texture
			var result = new TextureFrame(Crop.Width, Crop.Height) { InternalFormat = PixelInternalFormat.Rgba8, Wrap = TextureWrapMode.ClampToEdge, SequenceNumber = videoTexture.SequenceNumber };
			result.Alloc();

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, result.Texture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			// matrices
			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.Ortho(Crop.X, Crop.Right, VideoSize.Height - Crop.Bottom, VideoSize.Height - Crop.Y, 1, 10); // Bottom-left corner pixel has coordinate (0, 0)
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			// misc
			GL.Disable(EnableCap.DepthTest);
			GL.Disable(EnableCap.Lighting);
			GL.Enable(EnableCap.Texture2D);
			GL.Viewport(0, 0, Crop.Width, Crop.Height);
			GL.ClearColor(Color.Black);
			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

			// draw
			videoTexture.Bind();
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(0, 0); GL.Vertex3(0, 0, -8);
			GL.TexCoord2(1, 0); GL.Vertex3(VideoSize.Width, 0, -8);
			GL.TexCoord2(1, 1); GL.Vertex3(VideoSize.Width, VideoSize.Height, -8);
			GL.TexCoord2(0, 1); GL.Vertex3(0, VideoSize.Height, -8);
			GL.End();
			videoTexture.Unbind();

			// reset
			GL.Enable(EnableCap.DepthTest);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);

			//ReprojectWorldCheckerboardCorners();
			//ReprojectSRCheckerboardCorners();
			//ReprojectSRPixelCheckerboardCorners(distaFrame);

			return result;
		}

//=================================================================================================
		#endregion
		
		#region Reprojection test methods
//=================================================================================================
				
		/// <summary>Reprojects the world coordinates of the checkerboard corners.</summary>
		private void ReprojectWorldCheckerboardCorners()
		{
			// Step 1. Matrix setup: world to flea transformation ---------------------------------

			Matrix4 projMat = Utils.CreateCalibratedProjectionMatrix(
				VideoSize.Width, VideoSize.Height,
				Crop.X, Crop.Y, Crop.X + Crop.Width, Crop.Y + Crop.Height,
				TargetIntrinsics.Fx, TargetIntrinsics.Fy, TargetIntrinsics.Cx, TargetIntrinsics.Cy, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat.M12 *= -1;
			projMat.M22 *= -1;
			projMat.M32 *= -1;
			projMat.M42 *= -1;

			// negate Y and Z rows to convert from calibration to OpenGL coordinate system
			var viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW) * TargetExtrinsics.ToMatrix4();

			// OpenTK matrices are row-major, so transpose to get column-major data layout that OpenGL expects
			viewMat.Transpose();

			GL.MatrixMode(MatrixMode.Projection); GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadMatrix(ref viewMat);

			// Step 2. Reproject points -------------------------------------------------------------------

			GL.Disable(EnableCap.DepthTest);
			GL.PointSize(3);
			GL.Begin(BeginMode.Points);

			for (int x = 0; x < 8; x++)
			{
				for (int y = 0; y < 6; y++)
				{
					GL.Color3(x / 7f, 1f, y / 5f); // origin is green, x adds red, y adds blue
					GL.Vertex3(new Vector3(0.04f * x, 0.04f * y, 0f));
				}
			}

			GL.Color3(Color.White);
			GL.End();
			GL.Enable(EnableCap.DepthTest);
		}

		/// <summary>Reprojects the SR coordinates of the checkerboard corners.</summary>
		private void ReprojectSRCheckerboardCorners()
		{
			// Step 1. Matrix setup: world to flea transformation ---------------------------------

			Matrix4 projMat = Utils.CreateCalibratedProjectionMatrix(
				VideoSize.Width, VideoSize.Height,
				Crop.X, Crop.Y, Crop.X + Crop.Width, Crop.Y + Crop.Height,
				TargetIntrinsics.Fx, TargetIntrinsics.Fy, TargetIntrinsics.Cx, TargetIntrinsics.Cy, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat.M12 *= -1;
			projMat.M22 *= -1;
			projMat.M32 *= -1;
			projMat.M42 *= -1;

			var sourceView = SourceExtrinsics.ToMatrix4();
			var targetView = TargetExtrinsics.ToMatrix4();

			// undo the source view to get to world coordinates,
			// then apply the target view matrix
			sourceView.Invert();
			var viewMat = targetView * sourceView;

			// negate Y and Z rows to convert from calibration to OpenGL coordinate system
			viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW) * viewMat;

			// OpenTK matrices are row-major, so transpose to get column-major data layout that OpenGL expects
			viewMat.Transpose();

			GL.MatrixMode(MatrixMode.Projection); GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadMatrix(ref viewMat);

			// Step 2. Reproject points -------------------------------------------------------------------

			GL.Disable(EnableCap.DepthTest);
			GL.PointSize(3);
			GL.Begin(BeginMode.Points);

			var srView = SourceExtrinsics.ToMatrix4();
			srView.Transpose();
			for (int x = 0; x < 8; x++)
			{
				for (int y = 0; y < 6; y++)
				{
					GL.Color3(x / 7f, 1f, y / 5f); // origin is green, x adds red, y adds blue
					GL.Vertex3(Vector4.Transform(new Vector4(0.04f * x, 0.04f * y, 0f, 1f), srView).Xyz);
				}
			}

			GL.Color3(Color.White);
			GL.End();
			GL.Enable(EnableCap.DepthTest);
		}

		/// <summary>Reprojects the SR pixels closest to the projected checkerboard corners.</summary>
		private void ReprojectSRPixelCheckerboardCorners(BitmapFrame distaFrame)
		{
			// Step 1. Matrix setup: world to flea transformation ---------------------------------

			Matrix4 projMat = Utils.CreateCalibratedProjectionMatrix(
				VideoSize.Width, VideoSize.Height,
				Crop.X, Crop.Y, Crop.X + Crop.Width, Crop.Y + Crop.Height,
				TargetIntrinsics.Fx, TargetIntrinsics.Fy, TargetIntrinsics.Cx, TargetIntrinsics.Cy, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat.M12 *= -1;
			projMat.M22 *= -1;
			projMat.M32 *= -1;
			projMat.M42 *= -1;

			var sourceView = SourceExtrinsics.ToMatrix4();
			var targetView = TargetExtrinsics.ToMatrix4();

			// undo the source view to get to world coordinates,
			// then apply the target view matrix
			sourceView.Invert();
			var viewMat = targetView * sourceView;

			// negate Y and Z rows to convert from calibration to OpenGL coordinate system
			viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW) * viewMat;

			// OpenTK matrices are row-major, so transpose to get column-major data layout that OpenGL expects
			viewMat.Transpose();

			GL.MatrixMode(MatrixMode.Projection); GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadMatrix(ref viewMat);

			// Step 2. Reproject points -------------------------------------------------------------------

			GL.Disable(EnableCap.DepthTest);
			GL.PointSize(1);
			GL.Begin(BeginMode.Points);

			unsafe
			{
				fixed (byte* pByte = distaFrame.Data)
				{
					float* points = (float*)pByte;

					var srView = SourceExtrinsics.ToMatrix4(); srView.Transpose();
					var srCam  = SourceIntrinsics.ToMatrix4(); srCam.Transpose();

					for (int x = 0; x < 8; x++)
					{
						for (int y = 0; y < 6; y++)
						{
							for (int z = 0; z <= 8; z++)
							{
								// project checkerboard corners to find pixel coordinates
								var corner_world = new Vector4(0.04f * x, 0.04f * y, 0f, 1f);
								var corner_cam = Vector4.Transform(corner_world, srView);
								corner_cam.W *= (1 - z / 10f);
								var corner_proj = Vector4.Transform(corner_cam, srCam);
								var corner_pix = corner_proj.Xy / corner_proj.Z;

								// look at nearest pixel
								int px = (int)(corner_pix.X + 0.5) - (z / 3) + 1;
								int py = (int)(corner_pix.Y + 0.5) - (z % 3) + 1;
								var cx = points[py * distaFrame.Width * 4 + 4 * px];
								var cy = points[py * distaFrame.Width * 4 + 4 * px + 1];
								var cz = points[py * distaFrame.Width * 4 + 4 * px + 2];
								//var cd = points[py * distaFrame.Width * 4 + 4 * px + 3];

								GL.Color3(x / 7f, 1f, y / 5f); // origin is green, x adds red, y adds blue
								GL.Vertex3(-cx, -cy, cz);
							}
						}
					}

				}
			}

			GL.Color3(Color.White);
			GL.End();
			GL.Enable(EnableCap.DepthTest);
		}

//=================================================================================================
		#endregion

		#region Mesh drawing methods
//=================================================================================================

		private int VertexBufferID;
		private int IndexBufferID;
		private int NumVertices;
		private int NumIndices;

		private void SetupMeshMatrices()
		{
			if (SourceExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of source camera not provided.");
			if (TargetExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of target camera not provided.");
			if (TargetIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of target camera not provided.");


			// Step 1. Projection matrix from intrinsic parameters --------------------------------

			Matrix4 projMat = Utils.CreateCalibratedProjectionMatrix(
				VideoSize.Width, VideoSize.Height,
				Crop.X, Crop.Y, Crop.X + Crop.Width, Crop.Y + Crop.Height,
				TargetIntrinsics.Fx, TargetIntrinsics.Fy, TargetIntrinsics.Cx, TargetIntrinsics.Cy, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat.M12 *= -1;
			projMat.M22 *= -1;
			projMat.M32 *= -1;
			projMat.M42 *= -1;
			

			// Step 2. Modelview matrix to reproject view -----------------------------------------

			var sourceView = SourceExtrinsics.ToMatrix4();
			var targetView = TargetExtrinsics.ToMatrix4();

			// undo the source view to get to world coordinates,
			// then apply the target view matrix
			sourceView.Invert();
			var viewMat = targetView * sourceView;

			// negate Y and Z rows to convert from calibration to OpenGL coordinate system
			viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW) * viewMat;

			// OpenTK matrices are row-major, so transpose to get column-major data layout that OpenGL expects
			viewMat.Transpose();

			// original view
			//viewMat = Matrix4.Identity; // source view


			// Step 3. Set matrices ---------------------------------------------------------------

			GL.MatrixMode(MatrixMode.Projection); GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadMatrix(ref viewMat);
		}

		private TextureFrame DrawMesh(TextureFrame pointCloudTexture)
		{
			// allocate the output texture
			var result = new TextureFrame(Crop.Width, Crop.Height) { InternalFormat = PixelInternalFormat.Rgba32f, Wrap = TextureWrapMode.ClampToEdge, SequenceNumber = pointCloudTexture.SequenceNumber };
			result.Alloc();

			SetupMeshMatrices();

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, result.Texture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			GL.ClearColor(Color.Black);
			//GL.ClearColor(Color.FromArgb(0, Color.Black)); // results in spreading NaN points in STF
			GL.Viewport(0, 0, Crop.Width, Crop.Height);
			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

			// push the current Array Buffer state so we can restore it later
			GL.PushClientAttrib(ClientAttribMask.ClientVertexArrayBit);

			// use program and bind depth image to texture unit 0
			GL.UseProgram(MeshProgramID);
			pointCloudTexture.Bind(TextureUnit.Texture0);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "sobelThreshold"),  SobelThreshold);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "normalThreshold"), NormalThreshold);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "cleanupMode"), (int)CleanupMode);
	
			// sets the vertex pointer to the currently bound array describing how the data is stored
			GL.BindBuffer(BufferTarget.ArrayBuffer, VertexBufferID);
			GL.VertexPointer(3, VertexPointerType.Float, Vector3.SizeInBytes, IntPtr.Zero);
			GL.EnableClientState(ArrayCap.VertexArray);

			// draw the triangles in the element array buffer
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, IndexBufferID);
			GL.DrawElements(BeginMode.Triangles, NumIndices, DrawElementsType.UnsignedInt, IntPtr.Zero);

			// restore the state
			GL.UseProgram(0);
			pointCloudTexture.Unbind();
			GL.PopClientAttrib();

			return result;
		}

//=================================================================================================
		#endregion
	}
}
