// LibAvWrap -- .NET wrapper for the libav* libraries
// --------------------------------------------------
// By Christian Richardt <christian@richardt.name>
// (based on AvDn 0.7, www.informikon.com/tutorials/)
//
// $Id: CodecContext.h 436 2010-10-15 16:26:38Z cr333 $

#pragma once

using namespace System;
using namespace System::Text;
using namespace System::Runtime::InteropServices;

extern "C" {
#include "libavcodec/avcodec.h"
}

#include "Enums.h"
#include "Wrapper.h"

namespace LibAvWrap
{
	// forward declaration(s)
	ref class LibAvFrame;
	ref class LibAvPacket;
	ref class LibAvStream;

	///<summary>Wrapper for the AVCodecContext structure.</summary>
	public ref class LibAvCodecContext : Wrapper<AVCodecContext>
	{
// constructors and destructor --------------------------------------------------------------------

	internal:
		///<summary>Construct a codec context from a given pointer to AVCodecContext.</summary>
		LibAvCodecContext(AVCodecContext* p) : Wrapper<AVCodecContext>(p) { }

		///<summary>Construct a codec context from a given native pointer.</summary>
		LibAvCodecContext(IntPtr ptr) : Wrapper<AVCodecContext>(ptr) { }

		///<summary>Destructor.</summary>
		~LibAvCodecContext() { }

// member functions -------------------------------------------------------------------------------
	public:

		/////<summary>
		///// Decodes the samples pointed to by in_ptr into out_ptr.
		/////<Remarks>Returns the number of bytes written to out buffer</Remarks>
		/////</summary>
		//int DecodeAudio( IntPtr out_ptr, IntPtr in_ptr, int in_size );

		// TODO: error handling
		///<summary>Close the codec context.</summary>
		void Close() { avcodec_close(pointer); }
		
		///<summary>Decodes the packet payload into the frame.
		///<Remarks>If finished is not zero then decoding was complete</Remarks>
		///</summary>
		void DecodeVideo([Out]LibAvFrame^% f, LibAvPacket^ p, [Out]Int32% got_picture_ptr);

		///<summary>Encodes samples into the given buffer location and size.</summary>
		int EncodeAudio(IntPtr buf_ptr, int size, IntPtr samples);

		///<summary>Encodes a frame into the given buffer location and size.</summary>
		int EncodeVideo(IntPtr buf_ptr, int size, LibAvFrame^ f);

		///<summary>Sets the packet PKT_FLAG_KEY if the current coded_frame is a key frame.</summary>
		void HandlePacketFlagKey(LibAvPacket^ packet);

		///<summary>Finds and opens the decoder given the current CodecId.</summary>
		int OpenDecoder();

		///<summary>Finds and opens the encoder given the current CodecId.</summary>
		int OpenEncoder();

		///<summary>Sets the packet time stamp by rescaling using the the current coded_frame time stamp and the stream time_base.</summary>
		void SetPacketPts(LibAvPacket^ p, LibAvStream^ s);
		
		///<summary>Set the fundamental unit of time (in seconds) in terms of which frame timestamps are represented.</summary>
		void SetTimeBase(int n, int d) { pointer->time_base.num = n; pointer->time_base.den = d; }

// properties -------------------------------------------------------------------------------------

		///<summary>Bit rate (encoding: set by user).</summary>
		property int BitRate { void set(int rate); }

		///<summary>Frame width.</summary>
		property int Width { int get(); void set(int w); }

		///<summary>Frame height.</summary>
		property int Height { int get(); void set(int h); }

		///<summary>Number of pictures in a group of pictures, or 0 for intra_only (encoding: set by user).</summary>
		property int GopSize { int get(); void set(int s); }

		///<summary>Pixel format.</summary>
		property LibAvWrap::PixelFormat PixFmt { LibAvWrap::PixelFormat get(); void set(LibAvWrap::PixelFormat f); }

		///<summary>Maximum number of B-frames between non-B-frames.</summary>
		property int MaxBFrames {
			int get() { return pointer->max_b_frames; }
			void set(int v) { pointer->max_b_frames = v; }
		}

		///<summary>Macroblock decision mode (encoding: set by user).</summary>
		property int MbDecision { int get(); void set(int d); }

		///<summary>Codec ID.</summary>
		property LibAvWrap::CodecID CodecId {
			LibAvWrap::CodecID get() { return (LibAvWrap::CodecID)(int)pointer->codec_id; }
			void set(LibAvWrap::CodecID c) { pointer->codec_id = (::CodecID)c; }
		}

		///<summary>Samples per second.</summary>
		property int SampleRate { int get(); void set(int r); }

		///<summary>Number of audio channels.</summary>
		property int Channels { int get(); void set(int c); }

		///<summary>Codec type.</summary>
		property LibAvWrap::MediaType MediaType { void set(LibAvWrap::MediaType t); }

		///<summary>Codec flags.</summary>
		property LibAvWrap::CodecFlags Flags {
			LibAvWrap::CodecFlags get() { return (LibAvWrap::CodecFlags)pointer->flags; }
			void set(LibAvWrap::CodecFlags flags) { pointer->flags = (int)flags; }
		}
	};

	} // namespace LibAvWrap