﻿// $Id: FlyCaptureSource.cs 410 2010-10-04 12:39:30Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FlyCapture2Managed;
using VideoLib.Frames;
using System.Collections;

namespace VideoLib.FlyCapture
{
	public class FlyCaptureSource : Node
	{
		/// <summary>The sequence number of the most recently captured frame.</summary>
		public int SequenceNumber { get; protected set; }

		/// <summary>The desired resolution of the video input. Default: 640.</summary>
		public int Width { get; set; }

		/// <summary>The desired framerate of the video input. Default: 15 fps.</summary>
		public Framerates Framerate { get; set; }

		/// <summary>Video format used while capturing. Default: YUV422.</summary>
		public Formats CaptureFormat { get; set; }

		/// <summary>If true, outputs a strobe pulse for synchronising other cameras. Default: false.</summary>
		public bool Strobe { get; set; }

		/// <summary>If true, enables automatic exposure; otherwise uses <see cref="Exposure" /> for manual gain. Default: true.</summary>
		public bool AutoExposure { get; set; }

		/// <summary>Gets or sets the exposure in EV.</summary>
		/// <value>Value between -7.5f and 2.4f. Default: 0.0f.</value>
		public float Exposure { get; set; }

		/// <summary>If true, enables automatic gain; otherwise uses <see cref="Gain" /> for manual gain. Default: true.</summary>
		public bool AutoGain { get; set; }

		/// <summary>Gets or sets the gain in dB.</summary>
		/// <value>Value between 0.0f and around 24.0f. Default: 0.0f.</value>
		public float Gain { get; set; }

		/// <summary>Gets or sets the gamma value.</summary>
		/// <value>Value between 0.5f and around 4.0f. Default: 1.0f.</value>
		public float Gamma { get; set; }

		/// <summary>If true, enables automatic saturation; otherwise uses <see cref="Saturation" /> for manual saturation. Default: true.</summary>
		public bool AutoSaturation { get; set; }

		/// <summary>Gets or sets the saturation in percent.</summary>
		/// <value>Value between 0 and 400 percent. Default: 100.0f.</value>
		public float Saturation { get; set; }

		/// <summary>If true, enables automatic shutter time; otherwise uses <see cref="Shutter" /> for manual saturation. Default: true.</summary>
		public bool AutoShutter { get; set; }

		/// <summary>Gets or sets the shutter time in ms.</summary>
		/// <value>Value between 0 and 1000/fps. Default: auto.</value>
		public float Shutter { get; set; }

		/// <summary>If true, enables automatic white balance; otherwise uses <see cref="WhiteBalanceBlue" />
		/// and <see cref="WhiteBalanceRed" /> for manual white balance. Default: true.</summary>
		public bool AutoWhiteBalance { get; set; }

		/// <summary>Gets or sets the blue balance for manual white balance.</summary>
		/// <value>Value between 0 and 1024. Default: 512.</value>
		public int WhiteBalanceBlue { get; set; }

		/// <summary>Gets or sets the red balance for manual white balance.</summary>
		/// <value>Value between 0 and 1024. Default: 512.</value>
		public int WhiteBalanceRed { get; set; }

		/// <summary>Local instance of the camera.</summary>
		protected ManagedCamera Camera;

		/// <summary>Constructs a new video source from a Point Grey Research camera.</summary>
		public FlyCaptureSource()
			: base()
		{
			Width         = 640;
			Framerate     = Framerates._15;
			CaptureFormat = Formats.Yuv422;
			Strobe        = false;

			Gamma            = 1.0f; // there is no 'auto gamma'
			AutoExposure     = true; Exposure         = 0;
			AutoGain         = true; Gain             = 0;
			AutoSaturation   = true; Saturation       = 100;
			AutoShutter      = true; Shutter          = 30;
			AutoWhiteBalance = true; WhiteBalanceBlue = 512; WhiteBalanceRed  = 512;

			SequenceNumber = 0;
			AddOutputPin("out");
		}

		public override void Init()
		{
			ManagedBusManager busMgr = new ManagedBusManager();

			if (busMgr.GetNumOfCameras() == 0)
				throw new CameraNotFoundException();

			// connect to the first camera
			Camera = new ManagedCamera();
			Camera.Connect(busMgr.GetCameraFromIndex(0u));

			// set up (auto) exposure, gain, gamma and white balance
			Camera.SetProperty(new CameraProperty(PropertyType.AutoExposure) { onOff = true, autoManualMode = AutoExposure,     absControl = true, absValue = Exposure });
			Camera.SetProperty(new CameraProperty(PropertyType.Gain)         { onOff = true, autoManualMode = AutoGain,         absControl = true, absValue = Gain });
			Camera.SetProperty(new CameraProperty(PropertyType.Gamma)        { onOff = true, autoManualMode = false,            absControl = true, absValue = Gamma });
			Camera.SetProperty(new CameraProperty(PropertyType.Saturation)   { onOff = true, autoManualMode = AutoSaturation,   absControl = true, absValue = Saturation });
			Camera.SetProperty(new CameraProperty(PropertyType.Shutter)      { onOff = true, autoManualMode = AutoShutter,      absControl = true, absValue = Shutter });
			Camera.SetProperty(new CameraProperty(PropertyType.WhiteBalance) { onOff = true, autoManualMode = AutoWhiteBalance, valueA = (uint)WhiteBalanceRed, valueB = (uint)WhiteBalanceBlue });

			// set up desired video resolution, format and framerate
			Camera.SetProperty(new CameraProperty(PropertyType.FrameRate) { onOff = true, autoManualMode = false });
			Camera.SetVideoModeAndFrameRate(new VideoMode(Width, CaptureFormat), (FrameRate)Framerate);

			// enable strobe mode if required
			// NB. If enabled, this set up the camera to emit a strobe pulse on GPIO port 0.
			//     This is often used to synchronise other cameras to some master camera.
			Camera.SetTriggerMode(new TriggerMode { onOff = false });
			Camera.SetStrobe(new StrobeControl { onOff = Strobe, polarity = 1, source = 0 });

			// start capturing images
			Camera.StartCapture();
		}

		public unsafe override void Process()
		{
			// retrieve an image
			ManagedImage rawImage = new ManagedImage();
			Camera.RetrieveBuffer(rawImage);

			// At the time of writing this, the only pixel format that is directly understood by VideoLib is Grey8.
			// All others need to be converted to something else, such as RGB24.
			BitmapFrame frame = null;
			if (rawImage.pixelFormat == FlyCapture2Managed.PixelFormat.PixelFormatMono8)
			{
				// convert image to BitmapFrame
				frame = new BitmapFrame(new IntPtr(rawImage.data), (int)rawImage.cols, (int)rawImage.stride, (int)rawImage.rows,
					VideoLib.Frames.PixelFormat.Grey8) { SequenceNumber = SequenceNumber };
				rawImage.Dispose();
			}
			else
			{
				// convert pixel format to RGB24
				ManagedImage image = new ManagedImage();
				rawImage.Convert(FlyCapture2Managed.PixelFormat.PixelFormatBgr, image);
				rawImage.Dispose();

				// convert image to BitmapFrame
				frame = new BitmapFrame(new IntPtr(image.data), (int)image.cols, (int)image.stride, (int)image.rows,
					VideoLib.Frames.PixelFormat.BGR24) { SequenceNumber = SequenceNumber };
				image.Dispose();
			}

			PushFrame(frame);
			SequenceNumber++;
		}

		/// <summary>Request the node to update its status.</summary>
		protected override void UpdateStatusImpl()
		{
			// Finished is a final status, cannot change to anything else later
			if (Status == NodeStatus.Finished) return;

			// Stopped is a final status, cannot change to anything else later (for now)
			if (Status == NodeStatus.Stopped) return;

			// don't update status if node is currently Running, to avoid multiple threads running it simultanously
			if (Status == NodeStatus.Running) return;

			// don't update status if node is ready to run
			if (Status == NodeStatus.Ready) return;

			// always ready: trying not to fall behind and drop frames
			Status = NodeStatus.Ready;
		}

		protected void CleanUp()
		{
			if (Camera != null)
			{
				// Stop capturing images & disconnect
				Camera.StopCapture();
				Camera.Disconnect();
				Camera = null;
			}
		}

		protected override void Dispose(bool disposing)
		{
			CleanUp();
			base.Dispose(disposing);
		}
	}
}
