﻿// $Id: VideoMode.cs 262 2010-07-28 13:25:26Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VideoLib.FlyCapture
{
	public class VideoMode
	{
		public int Width;
		public int Height;
		public Formats Format;

		public VideoMode(int width, int height, Formats format)
		{
			Width = width;
			Height = height;
			Format = format;
		}

		public VideoMode(int width, Formats format)
		{
			Width = width;
			Height = (width / 4) * 3;
			Format = format;
		}

		/// <summary>Returns other video formats supported for the given video resolution.</summary>
		public IEnumerable<Formats> GetOtherFormats()
		{
			return modeMap.Where(x => (x.Value.Width == Width && x.Value.Height == Height && x.Value.Format != Format)).Select(x => x.Value.Format);
		}

		/// <summary>Returns other video resolutions supported for the given video format.</summary>
		public IEnumerable<int> GetOtherResolutions()
		{
			return modeMap.Where(x => (x.Value.Format == Format && (x.Value.Width != Width || x.Value.Height == Height))).Select(x => x.Value.Width);
		}

		public static IEnumerable<VideoMode> GetAllVideoModes()
		{
			return modeMap.Values;
		}

		/// <summary>Implicit conversion to FlyCapture's VideoMode</summary>
		/// <param name="videoMode">An instance of a wrapped video mode.</param>
		public static implicit operator FlyCapture2Managed.VideoMode(VideoMode videoMode)
		{
			var res = modeMap.Where(x => (x.Value.Width == videoMode.Width && x.Value.Height == videoMode.Height && x.Value.Format == videoMode.Format));
			if (res.Count() == 0)
				throw new InvalidVideoModeException(videoMode);
			return res.First().Key;
		}


		protected static Dictionary<FlyCapture2Managed.VideoMode, VideoLib.FlyCapture.VideoMode> modeMap = new Dictionary<FlyCapture2Managed.VideoMode, VideoMode>();

		/// <summary>The static constructor initialises a map of all defined video modes.</summary>
		static VideoMode()
		{
			// full list of video mode enums (except Format 7)
			modeMap[FlyCapture2Managed.VideoMode.VideoMode160x120Yuv444]   = new VideoMode( 160,  120, Formats.Yuv444);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode320x240Yuv422]   = new VideoMode( 320,  240, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode640x480Yuv411]   = new VideoMode( 640,  480, Formats.Yuv411);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode640x480Yuv422]   = new VideoMode( 640,  480, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode640x480Rgb]      = new VideoMode( 640,  480, Formats.Rgb);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode640x480Y8]       = new VideoMode( 640,  480, Formats.Grey8);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode640x480Y16]      = new VideoMode( 640,  480, Formats.Grey16);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode800x600Yuv422]   = new VideoMode( 800,  600, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode800x600Rgb]      = new VideoMode( 800,  600, Formats.Rgb);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode800x600Y8]       = new VideoMode( 800,  600, Formats.Grey8);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode800x600Y16]      = new VideoMode( 800,  600, Formats.Grey16);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1024x768Yuv422]  = new VideoMode(1024,  768, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1024x768Rgb]     = new VideoMode(1024,  768, Formats.Rgb);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1024x768Y8]      = new VideoMode(1024,  768, Formats.Grey8);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1024x768Y16]     = new VideoMode(1024,  768, Formats.Grey16);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1280x960Yuv422]  = new VideoMode(1280,  960, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1280x960Rgb]     = new VideoMode(1280,  960, Formats.Rgb);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1280x960Y8]      = new VideoMode(1280,  960, Formats.Grey8);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1280x960Y16]     = new VideoMode(1280,  960, Formats.Grey16);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1600x1200Yuv422] = new VideoMode(1600, 1200, Formats.Yuv422);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1600x1200Rgb]    = new VideoMode(1600, 1200, Formats.Rgb);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1600x1200Y8]     = new VideoMode(1600, 1200, Formats.Grey8);
			modeMap[FlyCapture2Managed.VideoMode.VideoMode1600x1200Y16]    = new VideoMode(1600, 1200, Formats.Grey16);
		}
	}

	public static class VideoModeUtils
	{
		/// <summary>Finds all supported video modes for a given framerate (according to camera).</summary>
		/// <param name="camera">Instance of a ManagedCamera.</param>
		/// <param name="frameRate">Given framerate</param>
		/// <returns>An IEnumerable of all supported video modes.</returns>
		internal static IEnumerable<VideoMode> SupportedVideoModes(this FlyCapture2Managed.ManagedCamera camera, FlyCapture2Managed.FrameRate frameRate)
		{
			return VideoMode.GetAllVideoModes().Where(x => camera.GetVideoModeAndFrameRateInfo(x, frameRate));
		}
	}
}
