﻿// $Id: DownsamplingFilter.cs 501 2010-11-17 22:02:13Z cr333 $
using System.Collections.Generic;
using System.ComponentModel;

namespace VideoLib.OpenGL
{
	/// <summary>Node that downsamples an input texture using nearest neighbours.</summary>
	public class DownsamplingFilter : FboNode
	{
		/// <summary>Gets or sets the downsampling factor.</summary>
		[Uniform(), Description("Downsampling factor.")]
		public int Factor { get; set; }

		/// <summary>Gets or sets the sample offset in the input image. Default: 0.</summary>
		[Uniform(), DefaultValue(0), Description("The offset of the sample in the input image.")]
		public int SampleOffset { get; set; }

		/// <summary>Initialises a new instance of the <see cref="DownsamplingFilter"/> class.</summary>
		public DownsamplingFilter()
		{
			// default settings
			Factor = 2;
			SampleOffset = 0;

			AddDefaultPins();
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.DownsamplingFilter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "input", Unit = 0 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate the output texture of the new smaller size, but otherwise identical to the input texture
			var outputWidth  = (inputs["in"].Width  - SampleOffset + Factor - 1) / Factor;
			var outputHeight = (inputs["in"].Height - SampleOffset + Factor - 1) / Factor;
			var resultTexture = new TextureFrame(outputWidth, outputHeight)
			{
				SequenceNumber = inputs["in"].SequenceNumber,
				InternalFormat = inputs["in"].InternalFormat,
				MinFilter      = inputs["in"].MinFilter,
				MagFilter      = inputs["in"].MagFilter,
				WrapS          = inputs["in"].WrapS,
				WrapT          = inputs["in"].WrapT
			};
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
