/*
This source code is part of the Floating Textures test program, which
is a derivative of the complete Floating Textures program containing
the offline version with prereconstructed geometry and prerecorded
video. 
For more information on Floating Textures please have a look at the article

"Floating Textures" 
by M. Eisemann, B. De Decker, M. Magnor, 
P. Bekaert, E. de Aguiar, N. Ahmed, C. Theobalt, A. Sellent
Computer Graphics Forum (Proc. Eurographics EG'08), vol. 27, no. 2, April 2008 


Copyright (C) 2008 Martin Eisemann

This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 3 of the License, or (at your option) any later 
version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program; if not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file ImageProcessing.frag
 *
 * \brief Collection of common image processing functions
 */


/**
 * \brief 1x3 linear filter with arbitrary orientation
 *
 * \param filter - the 1x3 filter mask, should be normalized beforehand
 * \param dxy - the size and direction of one step during the iteration
 * \param source - the source texture which is to be filtered
 * \param texCoord - the 2D texture coord, where the filter is centered
 *
 * \return
 * - the filtered color value
 */
vec4 imageFilter1D3(vec4[3] mask, vec2 dxy, 
		    sampler2D source, vec2 texCoord)
{
  texCoord = texCoord + dxy;

  vec4 color = vec4(0.0, 0.0, 0.0, 0.0);
  for(int i=0; i<3; i++){
    color += texture2D(source, texCoord) * mask[i];    
    texCoord -= dxy; // convolution is defined with mirrored mask
  }
  return color;
}



/**
 * \brief 1x5 linear filter with arbitrary orientation
 *
 * \param filter - the 1x5 filter mask, should be normalized beforehand
 * \param dxy - the size and direction of one step during the iteration
 * \param source - the source texture which is to be filtered
 * \param texCoord - the 2D texture coord, where the filter is centered
 *
 * \return
 * - the filtered color value
 */
vec4 imageFilter1D5(float[5] mask, vec2 dxy, 
		    sampler2D source, vec2 texCoord)
{
  texCoord = texCoord + 2.0 * dxy;

  vec4 color = vec4(0.0, 0.0, 0.0, 0.0);
  for(int i=0; i<5; i++){
    color += texture2D(source, texCoord) * mask[i];    
    texCoord -= dxy; // convolution is defined with mirrored mask
  }
  return color;
}


/**
 * \brief Converts a rgb color value to a gray value
 *
 * \param rgb - the input color, which is to be converted. vec4 because texture lookups usually
 *
 * \return
 * - Returns the corresponding gray value
 */
const vec3 rgb2gray_weights = vec3(0.2989, 0.5871, 0.114);
float rgb2gray(vec4 rgb)
{
  vec3 result = rgb.rgb * rgb2gray_weights;
  return (result.r + result.g + result.b);
}
