﻿// $Id: Utils.cs 536 2011-09-13 19:55:07Z cr333 $
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Vision;
using VideoLib.Frames;
using System.Reflection;
using System.IO;

namespace VideoLib.OpenGL
{
	/// <summary>Utilities to simplify using of OpenGL/OpenTK.</summary>
	public static class Utils
	{
		#region FBOs
//=================================================================================================

		/// <summary>Checks the status of the currently bound Framebuffer Object.</summary>
		/// <returns>True if framebuffer complete; else false.</returns>
		[Conditional("DEBUG")]
		public static void CheckFboStatus()
		{
			switch (GL.CheckFramebufferStatus(FramebufferTarget.Framebuffer))
			{
				case FramebufferErrorCode.FramebufferComplete:
					//Trace.WriteLine("FBO: The framebuffer is complete and valid for rendering.");
					return;

				case FramebufferErrorCode.FramebufferIncompleteAttachment:
					Trace.WriteLine("FBO: One or more attachment points are not framebuffer attachment complete. This could mean there’s no texture attached or the format isn’t renderable. For color textures this means the base format must be RGB or RGBA and for depth textures it must be a DEPTH_COMPONENT format. Other causes of this error are that the width or height is zero or the z-offset is out of range in case of render to volume.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteMissingAttachment:
					Trace.WriteLine("FBO: There are no attachments.");
					break;

				/* case  FramebufferErrorCode.GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT: 
						 Trace.WriteLine("FBO: An object has been attached to more than one attachment point.");
						 break;
					 */

				case FramebufferErrorCode.FramebufferIncompleteDimensionsExt:
					Trace.WriteLine("FBO: Attachments are of different size. All attachments must have the same width and height.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteFormatsExt:
					Trace.WriteLine("FBO: The color attachments have different format. All color attachments must have the same format.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteDrawBuffer:
					Trace.WriteLine("FBO: An attachment point referenced by GL.DrawBuffers() doesn’t have an attachment.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteReadBuffer:
					Trace.WriteLine("FBO: The attachment point referenced by GL.ReadBuffers() doesn’t have an attachment.");
					break;

				case FramebufferErrorCode.FramebufferUnsupported:
					Trace.WriteLine("FBO: This particular FBO configuration is not supported by the implementation.");
					break;

				default:
					Trace.WriteLine("FBO: Status unknown. (yes, this is really bad.)");
					break;
			}
		}

//=================================================================================================
		#endregion // FBOs

		#region Matrices
//=================================================================================================

		/// <summary>Converts this Matrix4d to a Matrix4.</summary>
		public static Matrix4 ToMatrix4(this Matrix4d m)
		{
			return new Matrix4(m.Row0.ToVector4(), m.Row1.ToVector4(), m.Row2.ToVector4(), m.Row3.ToVector4());
		}

		/// <summary>Converts this Matrix4 to a Matrix4d.</summary>
		public static Matrix4d ToMatrix4d(this Matrix4 m)
		{
			return new Matrix4d(m.Row0.ToVector4d(), m.Row1.ToVector4d(), m.Row2.ToVector4d(), m.Row3.ToVector4d());
		}

		/// <summary>Converts extrinsic parameters to a projection matrix.</summary>
		/// <param name="p">The extrinsic parameters.</param>
		public static Matrix4 ToMatrix4(this ExtrinsicParameters p)
		{
			return new Matrix4(p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8], p[9], p[10], p[11], 0, 0, 0, 1);
		}

		/// <summary>Converts intrinsic parameters to a camera matrix.</summary>
		/// <param name="p">The intrinsic parameters.</param>
		public static Matrix4 ToMatrix4(this IntrinsicParameters p)
		{
			return new Matrix4(p.Fx, 0, p.Cx, 0, 0, p.Fy, p.Cy, 0, 0, 0, 1, 0, 0, 0, 0, 0);
		}

		/// <summary>Creates an OpenGL projection matrix from intrinsic camera parameters.</summary>
		/// <param name="width">Width of the sensor, in pixels.</param>
		/// <param name="height">Height of the sensor, in pixels.</param>
		/// <param name="fx">Horizontal focal length, in pixels.</param>
		/// <param name="fy">Vertical focal length, in pixels.</param>
		/// <param name="cx">Horizontal position of the principal point, in pixels from the centre of the top-left pixel.</param>
		/// <param name="cy">Vertical position of the principal point, in pixels from the centre of the top-left pixel.</param>
		/// <param name="near">Position of the near clipping plane, in world units.</param>
		/// <param name="far">Position of the far clipping plane, in world units.</param>
		/// <returns></returns>
		public static Matrix4 CreateCalibratedProjectionMatrix(int width, int height, float fx, float fy, float cx, float cy, float near, float far)
		{
			// account for flip of y-axis between OpenGL and calibration coordinate systems
			cy = height - cy - 1;

			// compute the 4 sides of the frustum
			float left   = (       - cx - 0.5f) / fx * near;
			float right  = (width  - cx - 0.5f) / fx * near;
			float top    = (height - cy - 0.5f) / fy * near;
			float bottom = (       - cy - 0.5f) / fy * near;

			// Note: This matrix is ready for use in OpenGL (column-major), i.e. it is
			//       transposed relative to normal OpenTK matrices (which are row-major).
			return Matrix4.CreatePerspectiveOffCenter(left, right, bottom, top, near, far);
		}

		/// <summary>Creates an OpenGL projection matrix from intrinsic camera parameters and a cropping viewport.</summary>
		/// <param name="width">Width of the sensor, in pixels.</param>
		/// <param name="height">Height of the sensor, in pixels.</param>
		/// <param name="x1">The x-coordinate of the bottom-left point in the viewport.</param>
		/// <param name="y1">The y-coordinate of the bottom-left point in the viewport.</param>
		/// <param name="x2">The x-coordinate of the top-right point just outside the viewport.</param>
		/// <param name="y2">The y-coordinate of the top-right point just outside the viewport.</param>
		/// <param name="fx">Horizontal focal length, in pixels.</param>
		/// <param name="fy">Vertical focal length, in pixels.</param>
		/// <param name="cx">Horizontal position of the principal point, in pixels from the centre of the top-left pixel.</param>
		/// <param name="cy">Vertical position of the principal point, in pixels from the centre of the top-left pixel.</param>
		/// <param name="near">Position of the near clipping plane, in world units.</param>
		/// <param name="far">Position of the far clipping plane, in world units.</param>
		public static Matrix4 CreateCalibratedProjectionMatrix(int width, int height, int x1, int y1, int x2, int y2, float fx, float fy, float cx, float cy, float near, float far)
		{
			// account for flip of y-axis between OpenGL and calibration coordinate systems
			cy = height - cy - 1;

			// compute the 4 sides of the frustum
			float left   = (x1 - cx - 0.5f) / fx * near;
			float right  = (x2 - cx - 0.5f) / fx * near;
			float top    = (y2 - cy - 0.5f) / fy * near;
			float bottom = (y1 - cy - 0.5f) / fy * near;

			// Note: This matrix is ready for use in OpenGL (column-major), i.e. it is
			//       transposed relative to normal OpenTK matrices (which are row-major).
			return Matrix4.CreatePerspectiveOffCenter(left, right, bottom, top, near, far);
		}

//=================================================================================================
		#endregion // Matrices

		#region PixelFormat
//=================================================================================================

		/// <summary>Converts the given VideoLib PixelFormat to OpenTK PixelType.</summary>
		/// <param name="format">The VideoLib pixel format.</param>
		public static PixelType ToOpenGLPixelType(this VideoLib.Frames.PixelFormat format)
		{
			var datatype = format & VideoLib.Frames.PixelFormat.TypeMask;
			switch (datatype)
			{
				case VideoLib.Frames.PixelFormat.Byte:   return PixelType.UnsignedByte;
				case VideoLib.Frames.PixelFormat.Short:  return PixelType.UnsignedShort;
				case VideoLib.Frames.PixelFormat.Int:    return PixelType.UnsignedInt;
				case VideoLib.Frames.PixelFormat.Single: return PixelType.Float;
				//case VideoLib.Frames.PixelFormat.Double: break; // not supported
				default:
					throw new NotSupportedException(String.Format("Pixel datatype \"{0}\" is not supported.", datatype));
			}
		}

		/// <summary>Converts the given VideoLib PixelFormat to OpenTK PixelFormat.</summary>
		/// <param name="format">The VideoLib pixel format.</param>
		public static OpenTK.Graphics.OpenGL.PixelFormat ToOpenGLPixelFormat(this VideoLib.Frames.PixelFormat format)
		{
			var layout = format & VideoLib.Frames.PixelFormat.LayoutMask;
			switch (layout)
			{
				case VideoLib.Frames.PixelFormat.Grey: return OpenTK.Graphics.OpenGL.PixelFormat.Luminance;
				case VideoLib.Frames.PixelFormat.BGR:  return OpenTK.Graphics.OpenGL.PixelFormat.Bgr;
				case VideoLib.Frames.PixelFormat.RGB:  return OpenTK.Graphics.OpenGL.PixelFormat.Rgb; // untested
				case VideoLib.Frames.PixelFormat.BGRA: return OpenTK.Graphics.OpenGL.PixelFormat.Bgra;
				case VideoLib.Frames.PixelFormat.RGBA: return OpenTK.Graphics.OpenGL.PixelFormat.Rgba; // untested
				//case VideoLib.Frames.PixelFormat.ARGB: break; // not supported
				//case VideoLib.Frames.PixelFormat.ABGR: break; // not supported
				default:
					throw new NotSupportedException(String.Format("Pixel layout \"{0}\" is not supported.", layout));
			}
		}

		/// <summary>Converts an OpenTK PixelFormat to an incomplete VideoLib PixelFormat (type is missing).</summary>
		/// <param name="format">The OpenTK pixel format.</param>
		public static VideoLib.Frames.PixelFormat ToVideoLibPixelFormat(this OpenTK.Graphics.OpenGL.PixelFormat format)
		{
			switch (format)
			{
				case OpenTK.Graphics.OpenGL.PixelFormat.Bgr:       return VideoLib.Frames.PixelFormat.BGR;
				case OpenTK.Graphics.OpenGL.PixelFormat.Bgra:      return VideoLib.Frames.PixelFormat.BGRA | VideoLib.Frames.PixelFormat.Alpha;
				case OpenTK.Graphics.OpenGL.PixelFormat.Luminance: return VideoLib.Frames.PixelFormat.Grey;
				case OpenTK.Graphics.OpenGL.PixelFormat.Rgb:       return VideoLib.Frames.PixelFormat.RGB;
				case OpenTK.Graphics.OpenGL.PixelFormat.Rgba:      return VideoLib.Frames.PixelFormat.RGBA | VideoLib.Frames.PixelFormat.Alpha;
				default: throw new NotSupportedException(String.Format("Pixel format \"{0}\" is not supported.", format));
			}
		}

		/// <summary>Converts an OpenTK PixelType to an incomplete VideoLib PixelFormat (layout is missing).</summary>
		/// <param name="type">The OpenTK pixel type.</param>
		public static VideoLib.Frames.PixelFormat ToVideoLibPixelFormat(this OpenTK.Graphics.OpenGL.PixelType type)
		{
			switch (type)
			{
				case PixelType.Float:         return VideoLib.Frames.PixelFormat.Single;
				case PixelType.UnsignedByte:  return VideoLib.Frames.PixelFormat.Byte;
				case PixelType.UnsignedInt:   return VideoLib.Frames.PixelFormat.Int;
				case PixelType.UnsignedShort: return VideoLib.Frames.PixelFormat.Short;
				default: throw new NotSupportedException(String.Format("Pixel type \"{0}\" is not supported.", type));
			}
		}

//=================================================================================================
		#endregion // PixelFormat

		#region Shaders
//=================================================================================================

		/// <summary>Creates and compiles a shader from the given source.</summary>
		/// <param name="shaderType">Type of the shader.</param>
		/// <param name="source">The shader source.</param>
		/// <returns>The shader ID.</returns>
		public static int CreateShader(ShaderType shaderType, string source)
		{
			return CreateShader(shaderType, source, "unknown");
		}
			
		/// <summary>Creates and compiles a shader from the given source.</summary>
		/// <param name="shaderType">Type of the shader.</param>
		/// <param name="source">The shader source.</param>
		/// <param name="name">The shader name (e.g. filename or resource).</param>
		/// <returns>The shader ID.</returns>
		public static int CreateShader(ShaderType shaderType, string source, string name)
		{
			// create and compile shader
			var shader = GL.CreateShader(shaderType);
			GL.ShaderSource(shader, source);
			GL.CompileShader(shader);

			// test for errors
			int status_code;
			GL.GetShader(shader, ShaderParameter.CompileStatus, out status_code);
			if (status_code != 1)
				throw new ApplicationException(String.Format("In '{0}':\n{1}", name, GL.GetShaderInfoLog(shader)));

#if DEBUG
			// show warnings in debug mode
			var log = GL.GetShaderInfoLog(shader);
			if (log != null && log != "")
			{
				Console.WriteLine("Compile log for \"{0}\":", name);
				Console.WriteLine(log);
			}
#endif

			return shader;
		}

		/// <summary>Creates and compiles a shader from an embedded resource.</summary>
		/// <param name="assembly">The assembly that embeds the shaders.</param>
		/// <param name="resourceName">Name of the embedded resource. The assembly name prefix can be left out.</param>
		/// <param name="type">The type of the shader.</param>
		/// <returns>The shader ID.</returns>
		public static int CreateShaderFromResource(Assembly assembly, string resourceName, ShaderType type)
		{
			// First try to read the shader source from the given resource name, such as "Shader.frag".
			// Remember the name of the loaded resource, so that compile errors and warnings can refer to it.
			var name   = resourceName;
			var source = TryReadShaderSourceFromResource(assembly, resourceName);

			// If the resource was not found, then try prefixing the assembly name, e.g. "VideoLib.OpenGL.Shader.frag".
			if (source == null)
			{
				name   = String.Format("{0}.{1}", assembly.GetName().Name, resourceName);
				source = TryReadShaderSourceFromResource(assembly, name);
			}

			if (source == null)
				throw new ArgumentException(String.Format("Embedded resource '{0}' not found (also tried '{1}').", resourceName, name), "resourceName");

			return CreateShader(type, source, name);
		}

		/// <summary>Creates and compiles a shader from an embedded resource.</summary>
		/// <param name="resourceName">Complete name of the embedded resource.</param>
		/// <param name="type">The type of the shader.</param>
		/// <returns>The shader ID.</returns>
		public static int CreateShaderFromResource(string resourceName, ShaderType type)
		{
			return CreateShaderFromResource(Assembly.GetCallingAssembly(), resourceName, type);
		}

		/// <summary>Creates and compiles a shader from an embedded resource.</summary>
		/// <remarks>The shader type is guessed from the filename extension, so use '.frag', '.geom' or '.vert'.</remarks>
		/// <param name="resourceName">Complete name of the embedded resource.</param>
		/// <returns>The shader ID.</returns>
		public static int CreateShaderFromResource(string resourceName)
		{
			return CreateShaderFromResource(Assembly.GetCallingAssembly(), resourceName, GuessShaderType(resourceName));
		}

		/// <summary>Guesses the type of the shader from its name (frag, geom, vert).</summary>
		/// <param name="resourceName">Name of the resource.</param>
		/// <returns>The shader type guessed.</returns>
		private static ShaderType GuessShaderType(string resourceName)
		{
			if (resourceName.ToLower().EndsWith(".frag")) return ShaderType.FragmentShader;
			if (resourceName.ToLower().EndsWith(".geom")) return ShaderType.GeometryShader;
			if (resourceName.ToLower().EndsWith(".vert")) return ShaderType.VertexShader;

			throw new ArgumentException("Shader type cannot be guessed from name. Use .frag, .geom or .vert extensions.");
		}

		/// <summary>Attaches and links shaders to a program.</summary>
		/// <param name="program">The program.</param>
		/// <param name="shaders">The shaders.</param>
		public static int LinkShaderProgram(int program, params int[] shaders)
		{
			// attach all shaders
			foreach (var shader in shaders)
				GL.AttachShader(program, shader);

			// link all shaders
			GL.LinkProgram(program);

			// test for errors
			int status_code;
			GL.GetProgram(program, ProgramParameter.LinkStatus, out status_code);
			if (status_code != 1)
				throw new ApplicationException(GL.GetProgramInfoLog(program));

#if DEBUG
			// show warnings in debug mode
			var log = GL.GetProgramInfoLog(program);
			if (log != null && log != "")
			{
				Console.WriteLine("Link log:");
				Console.WriteLine(log);
			}
#endif

			// return program object
			return program;
		}
		
		/// <summary>Attaches and links shaders to a program.</summary>
		/// <param name="shaders">The shaders.</param>
		public static int LinkShaders(params int[] shaders)
		{
			int program = GL.CreateProgram();
			LinkShaderProgram(program, shaders);
			return program;
		}

		/// <summary>Attempts to read a shader source from an embedded resource.</summary>
		/// <param name="assembly">The assembly that embeds the shaders.</param>
		/// <param name="resourceName">Complete name of the embedded resource.</param>
		/// <returns>The shader source if found; otherwise null.</returns>
		public static string TryReadShaderSourceFromResource(Assembly assembly, string resourceName)
		{
			// try to read from resource
			var shaderStream = assembly.GetManifestResourceStream(resourceName);
			if (shaderStream == null) return null;

			// read the entire stream
			var streamReader = new StreamReader(shaderStream);
			var shaderSource = streamReader.ReadToEnd();
			streamReader.Dispose();

			return shaderSource;
		}

//=================================================================================================
		#endregion // Shaders

		#region Textures
//=================================================================================================

		/// <summary>Creates a new OpenGL texture from the given bitmap frame.</summary>
		/// <param name="bitmap">The bitmap frame.</param>
		/// <returns>The texture ID of the texture containing the bitmap.</returns>
		public static int CreateTexture(this BitmapFrame bitmap)
		{
			var tex = new Texture(bitmap.Width, bitmap.Height);
			return tex.Create(bitmap);
		}

		/// <summary>Decrements the reference counter of all textures in the dictionary.</summary>
		/// <param name="textures">Dictionary of textures.</param>
		public static void RefDec(this Dictionary<string, TextureFrame> textures)
		{
			foreach (var texture in textures)
				texture.Value.RefDec();
		}

		/// <summary>Decrements the reference counter of all textures.</summary>
		/// <param name="textures">IEnumerable of textures.</param>
		public static void RefDec(this IEnumerable<TextureFrame> textures)
		{
			foreach (var texture in textures)
				texture.RefDec();
		}

		/// <summary>Unbinds all textures in the dictionary.</summary>
		/// <param name="textures">Dictionary of textures.</param>
		public static void Unbind(this Dictionary<string, TextureFrame> textures)
		{
			foreach (var texture in textures)
				texture.Value.Unbind();
		}

		/// <summary>Unbinds all textures.</summary>
		/// <param name="textures">Dictionary of textures.</param>
		public static void Unbind(this IEnumerable<TextureFrame> textures)
		{
			foreach (var texture in textures)
				texture.Unbind();
		}

//=================================================================================================
		#endregion // textures

		#region Vectors
//=================================================================================================

		/// <summary>Converts this Vector2 to a float array.</summary>
		public static float[] ToFloatArray(this Vector2 v)
		{
			return new float[] { v.X, v.Y };
		}

		/// <summary>Converts this Vector3 to a float array.</summary>
		public static float[] ToFloatArray(this Vector3 v)
		{
			return new float[] { v.X, v.Y, v.Z };
		}

		/// <summary>Converts this Vector4 to a float array.</summary>
		public static float[] ToFloatArray(this Vector4 v)
		{
			return new float[] { v.X, v.Y, v.Z, v.W };
		}

		/// <summary>Converts this float array to a Vector2.</summary>
		public static Vector2 ToVector2(this float[] v)
		{
			if (v.Length != 2) throw new ArgumentException("Float array must have length 2.", "v");
			return new Vector2(v[0], v[1]);
		}

		/// <summary>Converts this Vector2d to a Vector2.</summary>
		public static Vector2 ToVector2(this Vector2d v)
		{
			return new Vector2((float)v.X, (float)v.Y);
		}

		/// <summary>Converts this Vector2 to a Vector2d.</summary>
		public static Vector2d ToVector2d(this Vector2 v)
		{
			return new Vector2d(v.X, v.Y);
		}

		/// <summary>Converts this float array to a Vector3.</summary>
		public static Vector3 ToVector3(this float[] v)
		{
			if (v.Length != 3) throw new ArgumentException("Float array must have length 3.", "v");
			return new Vector3(v[0], v[1], v[2]);
		}

		/// <summary>Converts this Vector3d to a Vector3.</summary>
		public static Vector3 ToVector3(this Vector3d v)
		{
			return new Vector3((float)v.X, (float)v.Y, (float)v.Z);
		}

		/// <summary>Converts this Vector3 to a Vector3d.</summary>
		public static Vector3d ToVector3d(this Vector3 v)
		{
			return new Vector3d(v.X, v.Y, v.Z);
		}

		/// <summary>Converts this float array to a Vector4.</summary>
		public static Vector4 ToVector4(this float[] v)
		{
			if (v.Length != 4) throw new ArgumentException("Float array must have length 4.", "v");
			return new Vector4(v[0], v[1], v[2], v[3]);
		}

		/// <summary>Converts this Vector4d to a Vector4.</summary>
		public static Vector4 ToVector4(this Vector4d v)
		{
			return new Vector4((float)v.X, (float)v.Y, (float)v.Z, (float)v.W);
		}

		/// <summary>Converts this Vector4 to a Vector4d.</summary>
		public static Vector4d ToVector4d(this Vector4 v)
		{
			return new Vector4d(v.X, v.Y, v.Z, v.W);
		}

//=================================================================================================
		#endregion // Vectors
	}
}
