// $Id: SwissRanger4000Source.cpp 262 2010-07-28 13:25:26Z cr333 $
#include "SwissRanger4000Source.h"

using namespace VideoLib::Frames;

namespace VideoLib { namespace SR4000 {

	SwissRanger4000Source::SwissRanger4000Source() : Node(), _srcam(NULL)
	{
		SequenceNumber = 0;

		SwissRanger4000Source::AcquireMode = AM::CorFixPtrn | AM::ConvGray | AM::DenoiseAnf;
		AutoExposure = true;
		IntegrationTime = 3;
		Coordinates = false;
		Timeout = 1000;

		// suggested settings for SR4000 (per SwissRanger manual)
		AEMinIntegrationTime = TimeToMs(1);
		AEMaxIntegrationTime = TimeToMs(150);
		AEPercentOverPos = 5;
		AEDesiredPos = 70;

		AddOutputPins("depth", "amplitude", "confidence", "coords");
	}

	SwissRanger4000Source::~SwissRanger4000Source()
	{
		// TODO: dispose managed resources

		// call finaliser
		this->!SwissRanger4000Source();
	}

    SwissRanger4000Source::!SwissRanger4000Source()
    {
		// TODO: free native resources
		Close();
    }

// Node interface ---------------------------------------------------------------------------------

	void SwissRanger4000Source::Init()
	{
		OpenUSB();

		// force all settings to be applied
		//AcquireMode = AcquireMode;
		ApplySettings(true);
	}

	void SwissRanger4000Source::Process()
	{
		if(!_srcam)
			throw gcnew Exception("SR4000 not initialised.");

		// acquire a frame
		Acquire();

		// push out the individual images
		for each(ImageEntry entry in GetImageList())
		{
			// debug: Console::WriteLine(String::Format("New image ({0}x{1}, {3}, {2})", entry.Width, entry.Height, entry.ImageType, entry.DataType));
			BitmapFrame^ frame = entry.GetBitmapFrame();
			frame->SequenceNumber = SequenceNumber;

			switch(entry.ImageType)
			{
				case ImageType::Distance:  PushFrame("depth",      frame); break;
				case ImageType::Amplitude: PushFrame("amplitude",  frame); break;
				case ImageType::ConfMap:   PushFrame("confidence", frame); break;
			}
		}

		// output the 3D coordinates as well, if requested
		if(Coordinates)
		{
			// initialise output frame
			BitmapFrame^ frame = gcnew BitmapFrame(Width, Height, PixelFormat::Single | PixelFormat::RGB);
			pin_ptr<unsigned char> pByte = &frame->Data[0];
			float* pFloat = (float*)pByte;

			// convert from spherical to Cartesian coordinates (interleave like XYZXYZ...)
			int res = SR_CoordTrfFlt(*_srcam, pFloat, &pFloat[1], &pFloat[2],
				3 * sizeof(float), 3 * sizeof(float), 3 * sizeof(float));

			// output the result
			frame->SequenceNumber = SequenceNumber;
			PushFrame("coords", frame);
		}

		SequenceNumber++;

		//System::Console::WriteLine(String::Format("Integration time : {0}", IntegrationTime));

		ApplySettings();
	}

	void SwissRanger4000Source::UpdateStatusImpl()
	{
		// Finished is a final status, cannot change to anything else later
		if (Status == NodeStatus::Finished) return;

		// Stopped is a final status, cannot change to anything else later (for now)
		if (Status == NodeStatus::Stopped) return;

		// don't update status if node is currently Running, to avoid multiple threads running it simultanously
		if (Status == NodeStatus::Running) return;

		// don't update status if node is ready to run
		if (Status == NodeStatus::Ready) return;

		// always ready: trying not to fall behind and drop frames
		Status = NodeStatus::Ready;
	}

// wrapped libMesaSR functions --------------------------------------------------------------------

	void SwissRanger4000Source::Acquire()
	{
		if(!_srcam)
			throw gcnew Exception("Cannot acquire images without an open camera.");

		SR_Acquire(*_srcam);
	}
	
	IEnumerable<ImageEntry>^ SwissRanger4000Source::GetImageList()
	{
		if(!_srcam)
			throw gcnew Exception("SR4000 not initialised.");

		// query for all image entries
		ImgEntry* imageEntryArray = NULL;
		int numImages = SR_GetImageList(*_srcam, &imageEntryArray);

		// then copy them into a managed list
		List<ImageEntry>^ imageEntries = gcnew List<ImageEntry>();
		for(int i = 0; i < numImages; i++)
			imageEntries->Add(ImageEntry(imageEntryArray[i]));

		return imageEntries;
	}

	void SwissRanger4000Source::OpenDialog(int mode, HWND parent)
	{
		// first try to close any open SR camera
		Close();

		// open dialog to select a camera (doesn't work? CR 2010-07-08)
		_srcam = new SRCAM;
		if(0 == SR_OpenDlg(_srcam, mode, parent))
			throw gcnew Exception("Could not connect to any SwissRanger SR4000 camera using the dialog.");
	}

	void SwissRanger4000Source::OpenUSB(unsigned int serialNumber)
	{
		// first try to close any open SR camera
		Close();

		_srcam = new SRCAM;
		if(0 == SR_OpenUSB(_srcam, serialNumber))
			throw gcnew Exception("Could not connect to any SwissRanger SR4000 camera via USB.");
	}

	void SwissRanger4000Source::Close()
	{
		if(_srcam != NULL)
		{
			switch(SR_Close(*_srcam))
			{
				case 0:  break;
				case 1:  throw gcnew Exception("Tried to close wrong device.");
				case 2:  throw gcnew Exception("Cannot release interface.");
				case 3:  throw gcnew Exception("Cannot close device.");
				default: throw gcnew Exception("Unknown error when closing device.");
			}
			delete _srcam;
			_srcam = NULL;
		}
	}

// additional functions ---------------------------------------------------------------------------

	void SwissRanger4000Source::ApplySettings(bool force)
	{
		if(_dirty || force)
		{
			if(!_srcam) throw gcnew Exception("Camera not initialised.");

			// apply acquire mode
			SR_SetMode(*_srcam, (int)_acquireMode);

			// apply auto exposure or set fixed integration time
			if(AutoExposure)
			{
				SR_SetAutoExposure(*_srcam, TimeFromMs(AEMinIntegrationTime),
					TimeFromMs(AEMaxIntegrationTime), AEPercentOverPos,	AEDesiredPos);
			}
			else
			{
				SR_SetAutoExposure(*_srcam, 0xff, 0, 0, 0);
				SR_SetIntegrationTime(*_srcam, TimeFromMs(_integrationTime));
			}

			// set timeout when communicating via USB or ethernet
			SR_SetTimeout(*_srcam, _timeout);

			_dirty = false;
		}
	}

	void SwissRanger4000Source::EnableMode(AM acquireMode)
	{
		AcquireMode = AcquireMode | acquireMode;
	}

	void SwissRanger4000Source::DisableMode(AM acquireMode)
	{
		AcquireMode = AcquireMode | ~acquireMode;
	}

	unsigned char SwissRanger4000Source::TimeFromMs(float milliseconds)
	{
		float fTime = 10.0f * (milliseconds - 0.3f);
		return (unsigned char)max(0.0f, min(255.0f, fTime));
	}

	float SwissRanger4000Source::TimeToMs(unsigned char time)
	{
		// convert time as per SwissRanger manual
		return 0.3f + 0.1f * time;
	}

// properties -------------------------------------------------------------------------------------

	int SwissRanger4000Source::Width::get()
	{
		if(_srcam) return SR_GetCols(*_srcam);
		else       return -1;
	}

	int SwissRanger4000Source::Height::get()
	{
		if(_srcam) return SR_GetRows(*_srcam);
		else       return -1;
	}

	bool SwissRanger4000Source::ConfidenceMap::get()
	{
		return (int)(SwissRanger4000Source::AcquireMode & AM::ConfMap) > 0;
	}

	void SwissRanger4000Source::ConfidenceMap::set(bool value)
	{
		if(value) EnableMode( AM::ConfMap);
		else      DisableMode(AM::ConfMap);
	}

	bool SwissRanger4000Source::Triggered::get()
	{
		return (int)(SwissRanger4000Source::AcquireMode & AM::Triggered) > 0;
	}

	void SwissRanger4000Source::Triggered::set(bool value)
	{
		if(value) EnableMode( AM::Triggered);
		else      DisableMode(AM::Triggered);
	}

}} // closing namespace VideoLib::SR4000