﻿// $Id: InputPin.cs 551 2012-02-15 18:09:32Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Threading;
using System.Diagnostics;

namespace VideoLib
{
	/// <summary>Implements an input pin of some node.</summary>
	public class InputPin : Pin
	{
		/// <summary>Buffer of frames.</summary>
		private Queue<Frame> buffer;

		/// <summary>Default buffer size to use for new input pins.</summary>
		public static int DefaultBufferSize { get; set; }

		/// <summary>The maximum size of the buffer. Zero means infinite.</summary>
		public int BufferSize { get; set; }

		/// <summary>True if data is available for consumption.</summary>
		public bool DataAvailable { get { return (buffer.Count != 0); } }

		/// <summary>Current status of this input Pin.</summary>
		public override PinStatus Status
		{
			// getting status is always up-to-date
			get
			{
				if (BufferSize != 0 && buffer.Count >= BufferSize && Node.Status != NodeStatus.Finished)
					return PinStatus.Blocked;
				else
					return PinStatus.Ready;
			}

			// setting the status of the parent (Pin), raises StatusChanged event
			protected set { base.Status = value; }
		}

		/// <summary>Static initialiser sets default buffer size.</summary>
		static InputPin()
		{
			DefaultBufferSize = 2;
		}

		/// <summary>Constructs an input pin.</summary>
		/// <param name="name">Name of this input pin.</param>
		/// <param name="node">Parent node of this input pin.</param>
		public InputPin(string name, INode node)
			: base(name, node)
		{
			buffer = new Queue<Frame>();
			BufferSize = DefaultBufferSize;
			Status = PinStatus.Ready;
			NewFrame += new NewFrameEventHandler(node.NewFrameListener);
		}

		/// <summary>Returns the first buffered frame or null.</summary>
		/// <returns>First buffered frame, or null if the buffer is empty.</returns>
		public Frames.Frame GetFrame()
		{
			Frame frame = null;
			lock (buffer)
			{
				frame = buffer.Dequeue();
			}
			UpdateStatus();
			lock (buffer)
			{
				if (buffer.Count > 0)
					OnNewFrame(buffer.Peek());
			}
			return frame;
		}

		/// <summary>Listener for new incoming frames (from other nodes' output pins).</summary>
		/// <param name="sender">Source of the frame.</param>
		/// <param name="e">New frame arguments.</param>
		public void NewFrameListener(object sender, NewFrameEventArgs e)
		{
			if (Verbose)
			{
				Debug.WriteLine(String.Format("{0}: {3} >>> {1} #{2} >>> {4}", Thread.CurrentThread.Name,
					e.Frame.GetType().Name, e.Frame.SequenceNumber, sender, this));
			}

			lock (buffer)
			{
				if (BufferSize != 0 && buffer.Count >= BufferSize)
				{
					//throw new Exception(String.Format("Exceeding buffer capacity of {0} at {1}.", BufferSize, this));
					Debug.WriteLine(String.Format("Exceeded buffer capacity at {0}: {1}/{2}", this, buffer.Count, BufferSize));
				}

				// store the incoming frame
				buffer.Enqueue(e.Frame);
			} // hold lock as short as possible, to reduce risk of deadlock

			// check for full buffers
			UpdateStatus();

			// only invoke new frame event for the first frame in the queue
			// TODO: check if this is correct
			if (buffer.Count <= 1)
				OnNewFrame(e.Frame);
		}

		/// <summary>Updates this input pin's status.</summary>
		public override void UpdateStatus()
		{
			Status = Status;
		}

		/// <summary>Returns a string representation of this InputPin.</summary>
		/// <returns>For example "LibAvCodecSource.in(7)" for pin "in" with 7 buffered frames.</returns>
		public override string ToString()
		{
			return String.Format("{0}({1})", base.ToString(), buffer.Count);
		}

		/// <summary>Clears the buffer.</summary>
		public void ClearBuffer()
		{
			buffer.Clear();
		}

	}
}
