﻿// $Id: WorkItem.cs 293 2010-08-04 16:26:33Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;

namespace VideoLib.Scheduling
{
	/// <summary>Abstract base class for work items processed by some schedulers.</summary>
	public abstract class WorkItem
	{
		/// <summary>Runs this work item.</summary>
		public abstract void Run();
	}


	/// <summary>A work item that processes a node.</summary>
	public class ProcessingWorkItem : WorkItem
	{
		/// <summary>Gets or sets the node to be processed.</summary>
		public INode Node { get; protected set; }

		/// <summary>Gets or sets the profiler to use. Default: null (no profiling).</summary>
		public IProfiler Profiler { get; set; }

		/// <summary>Initializes a new instance of the <see cref="ProcessingWorkItem"/> class.</summary>
		/// <param name="node">The node to be processed.</param>
		public ProcessingWorkItem(INode node) { Node = node; }

		/// <summary>Runs this work item.</summary>
		public override void Run()
		{
			//Console.WriteLine("{0}: processing {1}", Thread.CurrentThread.Name, Node);
			INode node = Node;
			lock (node)
			{
				if(Profiler != null) Profiler.MarkStarting(node);
				Node.PreProcess();
				Node.Process();
				Node.PostProcess();
				if (Profiler != null) Profiler.MarkEnding(node);
			}
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		public override string ToString()
		{
			return String.Format("Processing: {0}", Node);
		}
	}


	/// <summary>A work item that transports a frame from an ouput to an input pin.</summary>
	public class TransportWorkItem : WorkItem
	{
		/// <summary>Gets or sets the source pin.</summary>
		public OutputPin Source { get; protected set; }

		/// <summary>Gets or sets the destination pin.</summary>
		public InputPin Destination { get; protected set; }

		/// <summary>Gets or sets the event args.</summary>
		public NewFrameEventArgs EventArgs { get; protected set; }

		/// <summary>Initializes a new instance of the <see cref="TransportWorkItem"/> class.</summary>
		/// <param name="source">The source pin.</param>
		/// <param name="destination">The destination pin.</param>
		/// <param name="eventArgs">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		public TransportWorkItem(OutputPin source, InputPin destination, NewFrameEventArgs eventArgs)
		{
			Source = source;
			Destination = destination;
			EventArgs = eventArgs;
		}

		/// <summary>Runs this work item.</summary>
		public override void Run()
		{
			//Console.WriteLine("{0}: {3} >>> {1} #{2} >>> {4}", Thread.CurrentThread.Name,
			//    EventArgs.Frame.GetType().Name, EventArgs.Frame.SequenceNumber, Source, Destination);
			Destination.NewFrameListener(Source, EventArgs);
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		public override string ToString()
		{
			return String.Format("Transport: {0} => {1}", EventArgs.Frame.GetType().Name, Destination);
		}
	}


	/// <summary>A work item that executes a <see cref="System.Action"/>, which can be a <c>delegate() { ... }</c>.</summary>
	public class ActionWorkItem : WorkItem
	{
		/// <summary>Gets or sets the action to execute.</summary>
		public Action Action { get; protected set; }

		/// <summary>Initializes a new instance of the <see cref="ActionWorkItem"/> class.</summary>
		public ActionWorkItem(Action action) { Action = action; }

		/// <summary>Runs this work item.</summary>
		public override void Run() { Action.Invoke(); }
	}
}
