﻿// $Id: VideoDisplay.cs 386 2010-09-17 16:15:42Z cr333 $
using System;
using System.Drawing;
using VideoLib.Frames;
using VideoLib.Utils;
using System.Diagnostics;
using System.Collections.Generic;

namespace VideoLib.Sinks
{
	/// <summary>Generic video display node with support for a caption.</summary>
	public class VideoDisplay : Node
	{
		/// <summary>Gets or sets the video panel.</summary>
		protected IVideoControl VideoPanel { get; set; } // TODO: handle Disposed event properly

		/// <summary>Caption to display in this panel.</summary>
		protected string Title { get; set; }

		/// <summary>For displaying the frames per second arriving at the display.</summary>
		protected FrameRateCounter fps = new FrameRateCounter(0.5, 1.0);
		
		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Title); }
		}

		/// <summary>Indicates if the display rate should be limited to 25 fps, to avoid processing bottleneck.</summary>
		public Boolean LimitDisplayRate { get; set; }

		private long nextFrame = 0;

		/// <summary>Initialises a new instance of the <see cref="VideoDisplay"/> class.</summary>
		/// <param name="targetControl">The target control.</param>
		/// <param name="title">The video title.</param>
		public VideoDisplay(IVideoControl targetControl, string title)
			: base()
		{
			Title = title;
			LimitDisplayRate = true;
			VideoPanel = targetControl;
			VideoPanel.Disposed += new EventHandler(ControlDisposed);

			AddInputPin("in");
		}

		/// <summary>Event handler for handling disposing of the video control.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void ControlDisposed(object sender, EventArgs e)
		{
			VideoPanel = null;
			Status = NodeStatus.Stopped;
		}

		/// <summary>Processes a video frame.</summary>
		public override void Process()
		{
			// skip over this when not yet ready
			if (VideoPanel == null) return;

			// get latest frame
			Frame frame = GetInputPin("in").GetFrame();
			fps.NewFrame();

			// stop at the end of the stream
			if (frame is EndOfStreamFrame)
			{
				Status = NodeStatus.Finished;
				return;
			}

			if (LimitDisplayRate && Stopwatch.GetTimestamp() < nextFrame)
			{
				// skip frame if necessary
			}
			else
			{
				// prepare caption
				string text = String.Format("{0}\nFrame: {1}\nFPS: {2:0.0}", Title, frame.SequenceNumber, fps.FPS);

				// draw the new frame 
				VideoPanel.Draw(frame, text);

				// compute time of next frame to display
				if(LimitDisplayRate)
					nextFrame = Stopwatch.GetTimestamp() + Stopwatch.Frequency / 25;
			}

			// decrement reference count, if applicable
			IRefCountedFrame refCounted = frame as IRefCountedFrame;
			if (refCounted != null) refCounted.RefDec();
		}
	}
}
